# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import List, Optional, Union, Any, cast, Dict, TYPE_CHECKING
import uuid

from azure.core.credentials_async import AsyncTokenCredential
from azure.core.credentials import AzureKeyCredential
from azure.core.async_paging import AsyncItemPaged
from azure.core.polling import AsyncLROPoller
from azure.core.tracing.decorator import distributed_trace
from azure.core.tracing.decorator_async import distributed_trace_async

from .._generated.aio._client import PhoneNumbersClient as PhoneNumbersClientGen
from .._generated.models import (
    PhoneNumberSearchRequest,
    PhoneNumberCapabilitiesRequest,
    PhoneNumberPurchaseRequest,
    PhoneNumberType,
    OperatorInformationOptions,
    OperatorInformationRequest,
    OperatorInformationResult,
    PhoneNumberCapabilities,
    PhoneNumberCountry,
    PhoneNumberOffering,
    PhoneNumberLocality,
    PhoneNumberSearchResult,
    PurchasedPhoneNumber,
    PhoneNumberCapabilityType,
    PhoneNumberAreaCode,
    PhoneNumberBrowseCapabilitiesRequest,
    AvailablePhoneNumber,
    PhoneNumbersReservationPurchaseRequest,
    PhoneNumbersReservation,
    PhoneNumbersBrowseRequest,
    PhoneNumberAssignmentType,
    PhoneNumbersBrowseResult
)
from .._shared.auth_policy_utils import get_authentication_policy
from .._shared.utils import parse_connection_str
from .._version import SDK_MONIKER
from .._api_versions import DEFAULT_VERSION

if TYPE_CHECKING:
    from azure.core.polling import PollingMethod

_DEFAULT_POLLING_INTERVAL_IN_SECONDS = 2


class PhoneNumbersClient:
    """A client to interact with the AzureCommunicationService Phone Numbers gateway.

    This client provides operations to interact with the phone numbers service

    :param str endpoint:
        The endpoint url for Azure Communication Service resource.
    :param Union[AsyncTokenCredential, AzureKeyCredential] credential:
        The credential we use to authenticate against the service.
    :keyword api_version: Azure Communication Phone Number API version.
        The default value is "2025-06-01".
        Note that overriding this default value may result in unsupported behavior.
    :paramtype api_version: str
    :keyword accepted_language: The locale to display in the localized fields in responses. e.g.
        'en-US'. Default value is None.
    :paramtype accepted_language: str
    """

    def __init__(
        self, endpoint: str, credential: Union[AsyncTokenCredential, AzureKeyCredential], **kwargs: Any
    ) -> None:
        try:
            if not endpoint.lower().startswith("http"):
                endpoint = "https://" + endpoint
        except AttributeError as e:
            raise ValueError("Account URL must be a string.") from e

        if not credential:
            raise ValueError(
                "You need to provide account shared key to authenticate.")

        self._endpoint = endpoint
        self._accepted_language = kwargs.pop("accepted_language", None)
        self._api_version = kwargs.pop("api_version", DEFAULT_VERSION.value)
        self._phone_number_client = PhoneNumbersClientGen(
            self._endpoint,
            api_version=self._api_version,
            authentication_policy=get_authentication_policy(
                endpoint, credential, is_async=True),
            sdk_moniker=SDK_MONIKER,
            **kwargs
        )

    @classmethod
    def from_connection_string(cls, conn_str: str, **kwargs: Any) -> "PhoneNumbersClient":
        """Create PhoneNumbersClient from a Connection String.

        :param str conn_str:
            A connection string to an Azure Communication Service resource.
        :returns: Instance of PhoneNumbersClient.
        :rtype: ~azure.communication.phonenumbers.aio.PhoneNumbersClient
        """
        endpoint, access_key = parse_connection_str(conn_str)

        return cls(endpoint, AzureKeyCredential(access_key), **kwargs)

    @distributed_trace_async
    async def begin_purchase_phone_numbers(
            self,
            search_id: str,
            *,
            agree_to_not_resell: bool = False,
            continuation_token: Optional[str] = None,
            polling: Union[bool, "PollingMethod"] = True,
            polling_interval: int = _DEFAULT_POLLING_INTERVAL_IN_SECONDS,
            **kwargs: Any) -> AsyncLROPoller[None]:
        """Purchases phone numbers.

        :param search_id: The search id.
        :type search_id: str
        :keyword agree_to_not_resell: The agreement to not resell the phone numbers. Defaults to False if
         not provided.
        :paramtype agree_to_not_resell: bool
        :keyword continuation_token: A continuation token to restart a poller from a saved state.
        :paramtype continuation_token: str or None
        :keyword polling: Pass in True if you'd like the LROBasePolling polling method,
            False for no polling, or your own initialized polling object for a personal polling strategy.
        :paramtype polling: bool or ~azure.core.polling.PollingMethod
        :keyword polling_interval: Default waiting time (seconds) between two polls
            for LRO operations if no Retry-After header is present.
        :paramtype polling_interval: int
        :returns: A poller to wait on the purchase operation.
        :rtype: ~azure.core.polling.AsyncLROPoller[None]
        """
        purchase_request = PhoneNumberPurchaseRequest(
            search_id=search_id, agree_to_not_resell=agree_to_not_resell)

        return await self._phone_number_client.phone_numbers.begin_purchase_phone_numbers(
            body=purchase_request,
            polling_interval=polling_interval,
            continuation_token=continuation_token,
            polling=polling,
            **kwargs
        )

    @distributed_trace_async
    async def begin_release_phone_number(
        self,
        phone_number: str,
        *,
        continuation_token: Optional[str] = None,
        polling: Union[bool, "PollingMethod"] = True,
        polling_interval: int = _DEFAULT_POLLING_INTERVAL_IN_SECONDS,
        **kwargs: Any) -> AsyncLROPoller[None]:
        """Releases an purchased phone number.

        :param phone_number: Phone number to be released, e.g. +11234567890.
        :type phone_number: str
        :keyword continuation_token: A continuation token to restart a poller from a saved state.
        :paramtype continuation_token: str or None
        :keyword polling: Pass in True if you'd like the LROBasePolling polling method,
            False for no polling, or your own initialized polling object for a personal polling strategy.
        :paramtype polling: bool or ~azure.core.polling.PollingMethod
        :keyword polling_interval: Default waiting time (seconds) between two polls
            for LRO operations if no Retry-After header is present.
        :paramtype polling_interval: int
        :returns: A poller to wait on the release operation.
        :rtype: ~azure.core.polling.AsyncLROPoller[None]
        """
        return await self._phone_number_client.phone_numbers.begin_release_phone_number(
            phone_number,
            polling_interval=polling_interval,
            continuation_token=continuation_token,
            polling=polling,
            **kwargs
        )

    @distributed_trace_async
    async def begin_search_available_phone_numbers(
        self,
        country_code: str,
        phone_number_type: Union[PhoneNumberType, str],
        assignment_type: Union[PhoneNumberAssignmentType, str],
        capabilities: PhoneNumberCapabilities,
        *,
        area_code: Optional[str] = None,
        quantity: Optional[int] = None,
        continuation_token: Optional[str] = None,
        polling: Union[bool, "PollingMethod"] = True,
        polling_interval: int = _DEFAULT_POLLING_INTERVAL_IN_SECONDS,
        **kwargs: Any
    ) -> AsyncLROPoller[PhoneNumberSearchResult]:
        """Search for available phone numbers to purchase.

        :param country_code: The ISO 3166-2 country code, e.g. US.
        :type country_code: str
        :param phone_number_type: Required. The type of phone numbers to search for, e.g. geographic,
            or tollFree. Possible values include: "geographic", "tollFree".
        :type phone_number_type: str or ~azure.communication.phonenumbers.PhoneNumberType
        :param assignment_type: Required. The assignment type of the phone numbers to search for. A
            phone number can be assigned to a person, or to an application. Possible values include:
            "user", "application".
        :type assignment_type: str or
            ~azure.communication.phonenumbers.PhoneNumberAssignmentType
        :param capabilities: Required. Capabilities of a phone number.
        :type capabilities: ~azure.communication.phonenumbers.PhoneNumberCapabilities
        :keyword area_code: The area code of the desired phone number, e.g. 425. If not set,
            any area code could be used in the final search.
        :paramtype area_code: str or None
        :keyword quantity: The quantity of phone numbers in the search. Default is 1.
        :paramtype quantity: int or None
        :keyword continuation_token: A continuation token to restart a poller from a saved state.
        :paramtype continuation_token: str or None
        :keyword polling: Pass in True if you'd like the LROBasePolling polling method,
         False for no polling, or your own initialized polling object for a personal polling strategy.
        :paramtype polling: bool or ~azure.core.polling.PollingMethod
        :keyword polling_interval: Default waiting time (seconds) between two polls
            for LRO operations if no Retry-After header is present.
        :paramtype polling_interval: int
        :returns: A poller to wait on the search results.
        :rtype: ~azure.core.polling.AsyncLROPoller[~azure.communication.phonenumbers.PhoneNumberSearchResult]
        """
        search_request = PhoneNumberSearchRequest(
            phone_number_type=phone_number_type,
            assignment_type=assignment_type,
            capabilities=capabilities,
            area_code=area_code,
        )
        if quantity is not None:
            search_request.quantity = quantity

        return await self._phone_number_client.phone_numbers.begin_search_available_phone_numbers(
            country_code,
            search_request,
            polling_interval=polling_interval,
            continuation_token=continuation_token,
            polling=polling,
            **kwargs
        )

    @distributed_trace_async
    async def begin_update_phone_number_capabilities(
        self,
        phone_number: str,
        sms: Optional[Union[str, PhoneNumberCapabilityType]] = None,
        calling: Optional[Union[str, PhoneNumberCapabilityType]] = None,
        *,
        continuation_token: Optional[str] = None,
        polling: Union[bool, "PollingMethod"] = True,
        polling_interval: int = _DEFAULT_POLLING_INTERVAL_IN_SECONDS,
        **kwargs: Any
    ) -> AsyncLROPoller[PurchasedPhoneNumber]:
        """Updates the capabilities of a phone number.

        :param phone_number: The phone number id in E.164 format. The leading plus can be either + or
            encoded as %2B, e.g. +11234567890.
        :type phone_number: str
        :param sms: Capability value for SMS.
        :type sms: str or ~azure.communication.phonenumbers.PhoneNumberCapabilityType
        :param calling: Capability value for calling.
        :type calling: str or ~azure.communication.phonenumbers.PhoneNumberCapabilityType
        :keyword continuation_token: A continuation token to restart a poller from a saved state.
        :paramtype continuation_token: str or None
        :keyword polling: Pass in True if you'd like the LROBasePolling polling method,
            False for no polling, or your own initialized polling object for a personal polling strategy.
        :paramtype polling: bool or ~azure.core.polling.PollingMethod
        :keyword polling_interval: Default waiting time (seconds) between two polls
            for LRO operations if no Retry-After header is present.
        :paramtype polling_interval: int
        :returns: A poller to wait on the update operation.
        :rtype: ~azure.core.polling.AsyncLROPoller[~azure.communication.phonenumbers.PurchasedPhoneNumber]
        """
        capabilities_request = PhoneNumberCapabilitiesRequest(
            calling=calling, sms=sms)

        if not phone_number:
            raise ValueError("phone_number can't be empty")
        return await self._phone_number_client.phone_numbers.begin_update_capabilities(
            phone_number,
            body=capabilities_request,
            polling_interval=polling_interval,
            continuation_token=continuation_token,
            polling=polling,
            **kwargs
        )

    @distributed_trace_async
    async def get_purchased_phone_number(self, phone_number: str, **kwargs: Any) -> PurchasedPhoneNumber:
        """Gets the details of the given purchased phone number.

        :param phone_number: The purchased phone number whose details are to be fetched in E.164 format,
         e.g. +11234567890.
        :type phone_number: str
        :return: The details of the given purchased phone number.
        :rtype: ~azure.communication.phonenumbers.PurchasedPhoneNumber
        """
        return await self._phone_number_client.phone_numbers.get_by_number(phone_number, **kwargs)

    @distributed_trace
    def list_purchased_phone_numbers(
        self,
        *,
        skip: int = 0,
        top: int = 100,
        **kwargs: Any
    ) -> AsyncItemPaged[PurchasedPhoneNumber]:
        """Gets the list of all purchased phone numbers.

        Gets the list of all purchased phone numbers.

        :keyword skip: An optional parameter for how many entries to skip, for pagination purposes. The
         default value is 0. Default value is 0.
        :paramtype skip: int
        :keyword top: An optional parameter for how many entries to return, for pagination purposes.
         The default value is 100. Default value is 100.
        :paramtype top: int
        :return: An iterator like instance of PurchasedPhoneNumber
        :rtype:
         ~azure.core.async_paging.AsyncItemPaged[~azure.communication.phonenumbers.PurchasedPhoneNumber]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        return self._phone_number_client.phone_numbers.list_phone_numbers(skip=skip, top=top, **kwargs)

    @distributed_trace
    def list_available_countries(
        self,
        *,
        skip: int = 0,
        **kwargs: Any
    ) -> AsyncItemPaged[PhoneNumberCountry]:
        """Gets the list of supported countries.

        Gets the list of supported countries.

        :keyword skip: An optional parameter for how many entries to skip, for pagination purposes. The
         default value is 0. Default value is 0.
        :paramtype skip: int
        :return: An iterator like instance of PhoneNumberCountry
        :rtype:
         ~azure.core.async_paging.AsyncItemPaged[~azure.communication.phonenumbers.PhoneNumberCountry]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        return self._phone_number_client.phone_numbers.list_available_countries(
            skip=skip, accept_language=self._accepted_language, **kwargs
        )

    @distributed_trace
    def list_available_localities(
        self,
        country_code: str,
        *,
        phone_number_type: Optional[Union[PhoneNumberType, str]] = None,
        administrative_division: Optional[str] = None,
        skip: int = 0,
        **kwargs: Any
    ) -> AsyncItemPaged[PhoneNumberLocality]:
        """Gets the list of cities or towns with available phone numbers.

        Gets the list of cities or towns with available phone numbers.

        :param country_code: The ISO 3166-2 country/region code, e.g. US. Required.
        :type country_code: str
        :keyword phone_number_type: An optional parameter for the type of phone numbers, 
         e.g. geographic, tollFree, mobile. Default value is None.
        :paramtype phone_number_type: str or ~azure.communication.phonenumbers.PhoneNumberType
        :keyword administrative_division: An optional parameter for the name of the state or province
         in which to search for the area code. e.g. California. Default value is None.
        :paramtype administrative_division: str
        :keyword skip: An optional parameter for how many entries to skip, for pagination purposes. The
         default value is 0. Default value is 0.
        :paramtype skip: int
        :return: An iterator like instance of PhoneNumberLocality
        :rtype:
         ~azure.core.async_paging.AsyncItemPaged[~azure.communication.phonenumbers.PhoneNumberLocality]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        return self._phone_number_client.phone_numbers.list_available_localities(
            country_code,
            administrative_division=administrative_division,
            accept_language=self._accepted_language,
            phone_number_type=phone_number_type,
            skip=skip,
            **kwargs
        )

    @distributed_trace
    def list_available_offerings(
        self,
        country_code: str,
        *,
        phone_number_type: Optional[Union[PhoneNumberType, str]] = None,
        assignment_type: Optional[Union[PhoneNumberAssignmentType, str]] = None,
        skip: int = 0,
        **kwargs: Any
    ) -> AsyncItemPaged[PhoneNumberOffering]:
        """List available offerings of capabilities with rates for the given country/region.

        List available offerings of capabilities with rates for the given country/region.

        :param country_code: The ISO 3166-2 country/region code, e.g. US. Required.
        :type country_code: str
        :keyword phone_number_type: Filter by phone number type, e.g. geographic, tollFree, mobile.
         Default value is None.
        :paramtype phone_number_type: str or ~azure.communication.phonenumbers.PhoneNumberType
        :keyword assignment_type: Filter by assignmentType, e.g. User, Application. Known values are:
         "person" and "application". Default value is None.
        :paramtype assignment_type: str or ~azure.communication.phonenumbers.PhoneNumberAssignmentType
        :keyword skip: An optional parameter for how many entries to skip, for pagination purposes. The
         default value is 0. Default value is 0.
        :paramtype skip: int
        :return: An iterator like instance of PhoneNumberOffering
        :rtype:
         ~azure.core.async_paging.AsyncItemPaged[~azure.communication.phonenumbers.PhoneNumberOffering]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        return self._phone_number_client.phone_numbers.list_offerings(
            country_code,
            phone_number_type=phone_number_type,
            assignment_type=assignment_type,
            accept_language=self._accepted_language,
            skip=skip,
            **kwargs
        )

    @distributed_trace
    def list_available_area_codes(
        self,
        country_code: str,
        phone_number_type: Union[PhoneNumberType, str],
        *,
        administrative_division: Optional[str] = None,
        assignment_type: Optional[Union[PhoneNumberAssignmentType, str]] = None,
        locality: Optional[str] = None,
        skip: int = 0,
        **kwargs: Any
    ) -> AsyncItemPaged[PhoneNumberAreaCode]:
        """Gets the list of available area codes.

        :param country_code: The ISO 3166-2 country/region two letter code, e.g. US. Required.
        :type country_code: str
        :param phone_number_type: Filter by phone number type, e.g. geographic, tollFree, mobile.
         Required.
        :type phone_number_type: str or ~azure.communication.phonenumbers.PhoneNumberType
        :keyword assignment_type: Filter by assignmentType, e.g. User, Application. Known values are:
         "person" and "application". Default value is None.
        :paramtype assignment_type: str or ~azure.communication.phonenumbers.PhoneNumberAssignmentType
        :keyword locality: The name of locality in which to search for the area code. e.g. Seattle.
         This is required if the phone number type is Geographic. Default value is None.
        :paramtype locality: str
        :keyword administrative_division: The name of the state or province in which to search for the
         area code. e.g. California. Default value is None.
        :paramtype administrative_division: str
        :keyword skip: An optional parameter for how many entries to skip, for pagination purposes. The
         default value is 0. Default value is 0.
        :paramtype skip: int
        :return: An iterator like instance of PhoneNumberAreaCode
        :rtype: ~azure.core.paging.ItemPaged[~azure.communication.phonenumbers.PhoneNumberAreaCode]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        return self._phone_number_client.phone_numbers.list_area_codes(
            country_code,
            phone_number_type=phone_number_type,
            accept_language=self._accepted_language,
            assignment_type=assignment_type,
            locality=locality,
            administrative_division=administrative_division,
            skip=skip,
            **kwargs
        )

    @distributed_trace_async
    async def search_operator_information(
        self,
        phone_numbers: Union[str, List[str]],
        *,
        options: Optional[OperatorInformationOptions] = None,
        **kwargs: Any
    ) -> OperatorInformationResult:
        """Searches for operator information for a given list of phone numbers.

        :param phone_numbers: The phone number(s) whose operator information should be searched
        :type phone_numbers: str or list[str]
        :keyword options: Options to modify the search.  Please note: use of options can affect the cost of the search.
        :type options: OperatorInformationOptions
        :return: A search result containing operator information associated with the requested phone numbers
        :rtype: ~azure.communication.phonenumbers.OperatorInformationResult
        """
        if not isinstance(phone_numbers, list):
            phone_numbers = [phone_numbers]
        if options is None:
            options = OperatorInformationOptions(
                include_additional_operator_details=False)
        request = OperatorInformationRequest(
            phone_numbers=phone_numbers, options=options)
        return await self._phone_number_client.phone_numbers.operator_information_search(request, **kwargs)

    async def __aenter__(self) -> "PhoneNumbersClient":
        await self._phone_number_client.__aenter__()
        return self

    async def __aexit__(self, *args: "Any") -> None:
        await self.close()

    async def close(self) -> None:
        """Close the :class:
        `~azure.communication.phonenumbers.aio.PhoneNumbersClient` session.
        """
        await self._phone_number_client.__aexit__()

    @distributed_trace_async
    async def get_reservation(
        self, reservation_id: str, **kwargs: Any
    ) -> PhoneNumbersReservation:
        """Gets a reservation by its ID.

        Retrieves the reservation with the given ID, including all of the phone numbers associated with
        it.

        :param reservation_id: The id of the reservation. Required.
        :type reservation_id: str
        :return: PhoneNumbersReservation
        :rtype: ~azure.communication.phonenumbers.PhoneNumbersReservation
        """
        return await self._phone_number_client.phone_numbers.get_reservation(
            reservation_id, **kwargs
        )

    @distributed_trace
    def list_reservations(
        self, *, max_page_size: int = 100, **kwargs: Any
    ) -> AsyncItemPaged[PhoneNumbersReservation]:
        """Lists all reservations.

        Retrieves a paginated list of all phone number reservations. Note that the reservations will
        not be populated with the phone numbers associated with them.

        :keyword max_page_size: An optional parameter for how many entries to return, for pagination
         purposes. The default value is 100. Default value is 100.
        :paramtype max_page_size: int
        :return: An iterator like instance of PhoneNumbersReservationItem
        :rtype:
         ~azure.core.async_paging.AsyncItemPaged[~azure.communication.phonenumbers.PhoneNumbersReservation]
        """

        return self._phone_number_client.phone_numbers.list_reservations(
            max_page_size=max_page_size,
            **kwargs)

    @distributed_trace_async
    async def create_or_update_reservation(
        self, *,
        reservation_id: str,
        numbers_to_add: Optional[List[AvailablePhoneNumber]] = None,
        numbers_to_remove: Optional[List[str]] = None,
        **kwargs: Any
    ) -> PhoneNumbersReservation:
        """Creates or updates a reservation by its ID.

        Updates the reservation with the given ID if it exists; or creates a new one otherwise.
        The response will be the updated state of the reservation.
        Updating a reservation will extend the expiration time of the reservation to 15 minutes
        after the last change, up to a maximum of 2 hours from creation time.
        Partial success is possible, in which case the result will contain phone numbers with error status.


        :keyword reservation_id: The ID of the reservation. It must be a valid UUID. If a reservation,
         with that ID exists it will be updated; ortherwise a new reservation will be created.
        :keyword numbers_to_add: List of phone numbers to add to the reservation.
        :paramtype numbers_to_add: list[~azure.communication.phonenumbers.AvailablePhoneNumber]
        :keyword numbers_to_remove: List of phone number IDs to remove from the reservation.
        :paramtype numbers_to_remove: list[str]
        :return: The updated reservation
        :rtype: ~azure.communication.phonenumbers.PhoneNumbersReservation
        """
        if reservation_id is None:
            raise ValueError("reservation_id cannot be None")

        try:
            uuid.UUID(reservation_id)
        except ValueError as exc:
            raise ValueError("reservation_id must be in valid UUID format") from exc

        phone_numbers: Dict[str, Optional[AvailablePhoneNumber]] = {}
        if numbers_to_add:
            for number in numbers_to_add:
                if number.id is not None:
                    phone_numbers[number.id] = number
        if numbers_to_remove:
            for number_id in numbers_to_remove:
                phone_numbers[number_id] = None

        # Cast to satisfy type checker - merge-patch operations allow None values
        reservation = PhoneNumbersReservation(
            phone_numbers=cast(Optional[Dict[str, AvailablePhoneNumber]], phone_numbers))

        return await self._phone_number_client.phone_numbers.create_or_update_reservation(
            reservation_id, reservation, **kwargs
        )

    @distributed_trace_async
    async def delete_reservation(
        self, reservation_id: str, **kwargs: Any
    ) -> None:
        """Deletes a reservation by its ID.

        Deletes the reservation with the given ID. Any phone number in the reservation will be released
        and made available for others to purchase. Only reservations with 'active' status can be
        deleted.

        :param reservation_id: The id of the reservation. Required.
        :type reservation_id: str
        :return: None
        :rtype: None
        """
        return await self._phone_number_client.phone_numbers.delete_reservation(reservation_id, **kwargs)

    @distributed_trace_async
    async def begin_purchase_reservation(
        self,
        reservation_id: str,
        *,
        agree_to_not_resell: bool = False,
        continuation_token: Optional[str] = None,
        polling: Union[bool, "PollingMethod"] = True,
        polling_interval: int = _DEFAULT_POLLING_INTERVAL_IN_SECONDS,
        **kwargs: Any
    ) -> AsyncLROPoller[None]:
        """Starts the purchase of all phone numbers in the reservation.

        Starts a long running operation to purchase all of the phone numbers in the reservation.
        Purchase can only be started for active reservations that at least one phone number. If any of
        the phone numbers in the reservation is from a country where reselling is not permitted, do not
        resell agreement is required.

        The agreement to not resell is a legal requirement in some countries in order to purchase phone numbers.
        For more information on which countries require this agreement, please refer to this documentation:
        https://learn.microsoft.com/azure/communication-services/concepts/numbers/sub-eligibility-number-capability

        :param reservation_id: The id of the reservation. Required.
        :type reservation_id: str
        :keyword agree_to_not_resell: The agreement to not resell the phone numbers. Defaults to False if
         not provided.
        :paramtype agree_to_not_resell: bool
        :keyword continuation_token: A continuation token to restart a poller from a saved state.
        :paramtype continuation_token: str or None
        :keyword polling: Pass in True if you'd like the LROBasePolling polling method,
            False for no polling, or your own initialized polling object for a personal polling strategy.
        :paramtype polling: bool or ~azure.core.polling.PollingMethod
        :keyword polling_interval: Default waiting time (seconds) between two polls
            for LRO operations if no Retry-After header is present.
        :paramtype polling_interval: int
        :return: A poller to wait on the purchase operation.
        :rtype: ~azure.core.polling.AsyncLROPoller[None]
        """
        reservation_purchase_request = PhoneNumbersReservationPurchaseRequest(
            agree_to_not_resell=agree_to_not_resell)

        return await self._phone_number_client.phone_numbers.begin_purchase_reservation(
            reservation_id,
            reservation_purchase_request,
            polling_interval=polling_interval,
            continuation_token=continuation_token,
            polling=polling,
            **kwargs
        )

    @distributed_trace_async
    async def browse_available_phone_numbers(
            self,
            *,
            country_code: str,
            phone_number_type: Union[str, PhoneNumberType],
            sms_capability: Optional[Union[str, PhoneNumberCapabilityType]] = None,
            calling_capability: Optional[Union[str, PhoneNumberCapabilityType]] = None,
            assignment_type: Optional[Union[str, PhoneNumberAssignmentType]] = None,
            phone_number_prefixes: Optional[List[str]] = None,
            **kwargs: Any,
    ) -> PhoneNumbersBrowseResult:
        """Browses for available phone numbers to purchase.

        Browses for available phone numbers to purchase. The response will be a randomized list of
        phone numbers available to purchase matching the browsing criteria. This operation is not
        paginated. Since the results are randomized, repeating the same request will not guarantee the
        same results.

        :keyword country_code: The ISO 3166-2 country code, e.g. US. Required.
        :paramtype country_code: str
        :keyword phone_number_type: Required. The type of phone numbers to search for, e.g. geographic,
            or tollFree. Possible values include: "geographic", "tollFree".
        :paramtype phone_number_type: str or ~azure.communication.phonenumbers.PhoneNumberType
        :keyword sms_capability: The SMS capability to search for. Known values are: "inbound",
            "outbound", "inbound_outbound", "none".
        :paramtype sms_capability: str or ~azure.communication.phonenumbers.PhoneNumberCapabilityType
        :keyword calling_capability: The calling capability to search for. Known values are: "inbound",
            "outbound", "inbound_outbound", "none".
        :paramtype calling_capability: str or ~azure.communication.phonenumbers.PhoneNumberCapabilityType
        :keyword assignment_type: Represents the assignment type of the offering. Also known as the use
            case. Known values are: "person" and "application".
        :paramtype assignment_type: str or ~azure.communication.phonenumbers.PhoneNumberAssignmentType
        :keyword phone_number_prefixes: The phone number prefix to match. If specified, the search will
            be limited to phone numbers that start with any of the given prefixes.
        :paramtype phone_number_prefixes: list[str]
        :return: A list of available phone numbers matching the browsing criteria.
        :rtype: ~azure.communication.phonenumbers.models.PhoneNumbersBrowseResult
        """
        if not country_code:
            raise ValueError("country_code is required.")
        if not phone_number_type:
            raise ValueError("phone_number_type is required.")

        browse_capabilities = None
        if sms_capability is not None or calling_capability is not None:
            browse_capabilities = PhoneNumberBrowseCapabilitiesRequest(
                calling=calling_capability,
                sms=sms_capability,
            )
        browse_request = PhoneNumbersBrowseRequest(
            phone_number_type=phone_number_type,
            capabilities=browse_capabilities,
            assignment_type=assignment_type,
            phone_number_prefixes=phone_number_prefixes
        )
        return await self._phone_number_client.phone_numbers.browse_available_numbers(
            country_code,
            browse_request,
            **kwargs
        )



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
import pytest
from azure.mgmt.cosmosdb import CosmosDBManagementClient

from devtools_testutils import AzureMgmtRecordedTestCase, RandomNameResourceGroupPreparer, recorded_by_proxy

AZURE_LOCATION = "eastus"


@pytest.mark.skip("you may need to update the auto-generated test case before run it")
class TestCosmosDBManagementCassandraResourcesOperations(AzureMgmtRecordedTestCase):
    def setup_method(self, method):
        self.client = self.create_mgmt_client(CosmosDBManagementClient)

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_get_cassandra_keyspace_throughput(self, resource_group):
        response = self.client.cassandra_resources.get_cassandra_keyspace_throughput(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_update_cassandra_keyspace_throughput(self, resource_group):
        response = self.client.cassandra_resources.begin_update_cassandra_keyspace_throughput(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            update_throughput_parameters={
                "properties": {
                    "resource": {
                        "autoscaleSettings": {
                            "maxThroughput": 0,
                            "autoUpgradePolicy": {"throughputPolicy": {"incrementPercent": 0, "isEnabled": bool}},
                            "targetMaxThroughput": 0,
                        },
                        "instantMaximumThroughput": "str",
                        "minimumThroughput": "str",
                        "offerReplacePending": "str",
                        "softAllowedMaximumThroughput": "str",
                        "throughput": 0,
                        "throughputBuckets": [{"id": 0, "maxThroughputPercentage": 0, "isDefaultBucket": bool}],
                    }
                },
                "id": "str",
                "identity": {
                    "principalId": "str",
                    "tenantId": "str",
                    "type": "str",
                    "userAssignedIdentities": {"str": {"clientId": "str", "principalId": "str"}},
                },
                "location": "str",
                "name": "str",
                "tags": {"str": "str"},
                "type": "str",
            },
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_migrate_cassandra_keyspace_to_autoscale(self, resource_group):
        response = self.client.cassandra_resources.begin_migrate_cassandra_keyspace_to_autoscale(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_migrate_cassandra_keyspace_to_manual_throughput(self, resource_group):
        response = self.client.cassandra_resources.begin_migrate_cassandra_keyspace_to_manual_throughput(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_get_cassandra_table_throughput(self, resource_group):
        response = self.client.cassandra_resources.get_cassandra_table_throughput(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            table_name="str",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_update_cassandra_table_throughput(self, resource_group):
        response = self.client.cassandra_resources.begin_update_cassandra_table_throughput(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            table_name="str",
            update_throughput_parameters={
                "properties": {
                    "resource": {
                        "autoscaleSettings": {
                            "maxThroughput": 0,
                            "autoUpgradePolicy": {"throughputPolicy": {"incrementPercent": 0, "isEnabled": bool}},
                            "targetMaxThroughput": 0,
                        },
                        "instantMaximumThroughput": "str",
                        "minimumThroughput": "str",
                        "offerReplacePending": "str",
                        "softAllowedMaximumThroughput": "str",
                        "throughput": 0,
                        "throughputBuckets": [{"id": 0, "maxThroughputPercentage": 0, "isDefaultBucket": bool}],
                    }
                },
                "id": "str",
                "identity": {
                    "principalId": "str",
                    "tenantId": "str",
                    "type": "str",
                    "userAssignedIdentities": {"str": {"clientId": "str", "principalId": "str"}},
                },
                "location": "str",
                "name": "str",
                "tags": {"str": "str"},
                "type": "str",
            },
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_migrate_cassandra_table_to_autoscale(self, resource_group):
        response = self.client.cassandra_resources.begin_migrate_cassandra_table_to_autoscale(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            table_name="str",
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_migrate_cassandra_table_to_manual_throughput(self, resource_group):
        response = self.client.cassandra_resources.begin_migrate_cassandra_table_to_manual_throughput(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            table_name="str",
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_get_cassandra_view_throughput(self, resource_group):
        response = self.client.cassandra_resources.get_cassandra_view_throughput(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            view_name="str",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_update_cassandra_view_throughput(self, resource_group):
        response = self.client.cassandra_resources.begin_update_cassandra_view_throughput(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            view_name="str",
            update_throughput_parameters={
                "properties": {
                    "resource": {
                        "autoscaleSettings": {
                            "maxThroughput": 0,
                            "autoUpgradePolicy": {"throughputPolicy": {"incrementPercent": 0, "isEnabled": bool}},
                            "targetMaxThroughput": 0,
                        },
                        "instantMaximumThroughput": "str",
                        "minimumThroughput": "str",
                        "offerReplacePending": "str",
                        "softAllowedMaximumThroughput": "str",
                        "throughput": 0,
                        "throughputBuckets": [{"id": 0, "maxThroughputPercentage": 0, "isDefaultBucket": bool}],
                    }
                },
                "id": "str",
                "identity": {
                    "principalId": "str",
                    "tenantId": "str",
                    "type": "str",
                    "userAssignedIdentities": {"str": {"clientId": "str", "principalId": "str"}},
                },
                "location": "str",
                "name": "str",
                "tags": {"str": "str"},
                "type": "str",
            },
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_migrate_cassandra_view_to_autoscale(self, resource_group):
        response = self.client.cassandra_resources.begin_migrate_cassandra_view_to_autoscale(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            view_name="str",
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_migrate_cassandra_view_to_manual_throughput(self, resource_group):
        response = self.client.cassandra_resources.begin_migrate_cassandra_view_to_manual_throughput(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            view_name="str",
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_get_cassandra_keyspace(self, resource_group):
        response = self.client.cassandra_resources.get_cassandra_keyspace(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_create_update_cassandra_keyspace(self, resource_group):
        response = self.client.cassandra_resources.begin_create_update_cassandra_keyspace(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            create_update_cassandra_keyspace_parameters={
                "properties": {
                    "resource": {"id": "str"},
                    "options": {"autoscaleSettings": {"maxThroughput": 0}, "throughput": 0},
                },
                "id": "str",
                "identity": {
                    "principalId": "str",
                    "tenantId": "str",
                    "type": "str",
                    "userAssignedIdentities": {"str": {"clientId": "str", "principalId": "str"}},
                },
                "location": "str",
                "name": "str",
                "tags": {"str": "str"},
                "type": "str",
            },
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_delete_cassandra_keyspace(self, resource_group):
        response = self.client.cassandra_resources.begin_delete_cassandra_keyspace(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_list_cassandra_keyspaces(self, resource_group):
        response = self.client.cassandra_resources.list_cassandra_keyspaces(
            resource_group_name=resource_group.name,
            account_name="str",
        )
        result = [r for r in response]
        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_get_cassandra_table(self, resource_group):
        response = self.client.cassandra_resources.get_cassandra_table(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            table_name="str",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_create_update_cassandra_table(self, resource_group):
        response = self.client.cassandra_resources.begin_create_update_cassandra_table(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            table_name="str",
            create_update_cassandra_table_parameters={
                "properties": {
                    "resource": {
                        "id": "str",
                        "analyticalStorageTtl": 0,
                        "defaultTtl": 0,
                        "schema": {
                            "clusterKeys": [{"name": "str", "orderBy": "str"}],
                            "columns": [{"name": "str", "type": "str"}],
                            "partitionKeys": [{"name": "str"}],
                        },
                    },
                    "options": {"autoscaleSettings": {"maxThroughput": 0}, "throughput": 0},
                },
                "id": "str",
                "identity": {
                    "principalId": "str",
                    "tenantId": "str",
                    "type": "str",
                    "userAssignedIdentities": {"str": {"clientId": "str", "principalId": "str"}},
                },
                "location": "str",
                "name": "str",
                "tags": {"str": "str"},
                "type": "str",
            },
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_delete_cassandra_table(self, resource_group):
        response = self.client.cassandra_resources.begin_delete_cassandra_table(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            table_name="str",
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_list_cassandra_tables(self, resource_group):
        response = self.client.cassandra_resources.list_cassandra_tables(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
        )
        result = [r for r in response]
        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_get_cassandra_view(self, resource_group):
        response = self.client.cassandra_resources.get_cassandra_view(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            view_name="str",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_create_update_cassandra_view(self, resource_group):
        response = self.client.cassandra_resources.begin_create_update_cassandra_view(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            view_name="str",
            create_update_cassandra_view_parameters={
                "properties": {
                    "resource": {"id": "str", "viewDefinition": "str"},
                    "options": {"autoscaleSettings": {"maxThroughput": 0}, "throughput": 0},
                },
                "id": "str",
                "identity": {
                    "principalId": "str",
                    "tenantId": "str",
                    "type": "str",
                    "userAssignedIdentities": {"str": {"clientId": "str", "principalId": "str"}},
                },
                "location": "str",
                "name": "str",
                "tags": {"str": "str"},
                "type": "str",
            },
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_delete_cassandra_view(self, resource_group):
        response = self.client.cassandra_resources.begin_delete_cassandra_view(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
            view_name="str",
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_list_cassandra_views(self, resource_group):
        response = self.client.cassandra_resources.list_cassandra_views(
            resource_group_name=resource_group.name,
            account_name="str",
            keyspace_name="str",
        )
        result = [r for r in response]
        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_get_cassandra_role_definition(self, resource_group):
        response = self.client.cassandra_resources.get_cassandra_role_definition(
            resource_group_name=resource_group.name,
            account_name="str",
            role_definition_id="str",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_create_update_cassandra_role_definition(self, resource_group):
        response = self.client.cassandra_resources.begin_create_update_cassandra_role_definition(
            resource_group_name=resource_group.name,
            account_name="str",
            role_definition_id="str",
            create_update_cassandra_role_definition_parameters={
                "id": "str",
                "name": "str",
                "properties": {
                    "assignableScopes": ["str"],
                    "id": "str",
                    "permissions": [{"dataActions": ["str"], "id": "str", "notDataActions": ["str"]}],
                    "roleName": "str",
                    "type": "str",
                },
                "systemData": {
                    "createdAt": "2020-02-20 00:00:00",
                    "createdBy": "str",
                    "createdByType": "str",
                    "lastModifiedAt": "2020-02-20 00:00:00",
                    "lastModifiedBy": "str",
                    "lastModifiedByType": "str",
                },
                "type": "str",
            },
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_delete_cassandra_role_definition(self, resource_group):
        response = self.client.cassandra_resources.begin_delete_cassandra_role_definition(
            resource_group_name=resource_group.name,
            account_name="str",
            role_definition_id="str",
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_list_cassandra_role_definitions(self, resource_group):
        response = self.client.cassandra_resources.list_cassandra_role_definitions(
            resource_group_name=resource_group.name,
            account_name="str",
        )
        result = [r for r in response]
        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_get_cassandra_role_assignment(self, resource_group):
        response = self.client.cassandra_resources.get_cassandra_role_assignment(
            resource_group_name=resource_group.name,
            account_name="str",
            role_assignment_id="str",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_create_update_cassandra_role_assignment(self, resource_group):
        response = self.client.cassandra_resources.begin_create_update_cassandra_role_assignment(
            resource_group_name=resource_group.name,
            account_name="str",
            role_assignment_id="str",
            create_update_cassandra_role_assignment_parameters={
                "id": "str",
                "name": "str",
                "properties": {
                    "principalId": "str",
                    "provisioningState": "str",
                    "roleDefinitionId": "str",
                    "scope": "str",
                },
                "systemData": {
                    "createdAt": "2020-02-20 00:00:00",
                    "createdBy": "str",
                    "createdByType": "str",
                    "lastModifiedAt": "2020-02-20 00:00:00",
                    "lastModifiedBy": "str",
                    "lastModifiedByType": "str",
                },
                "type": "str",
            },
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_begin_delete_cassandra_role_assignment(self, resource_group):
        response = self.client.cassandra_resources.begin_delete_cassandra_role_assignment(
            resource_group_name=resource_group.name,
            account_name="str",
            role_assignment_id="str",
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy
    def test_cassandra_resources_list_cassandra_role_assignments(self, resource_group):
        response = self.client.cassandra_resources.list_cassandra_role_assignments(
            resource_group_name=resource_group.name,
            account_name="str",
        )
        result = [r for r in response]
        # please add some check logic here by yourself
        # ...



# pylint: disable=line-too-long,useless-suppression
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from azure.identity import DefaultAzureCredential

from azure.mgmt.authorization import AuthorizationManagementClient

"""
# PREREQUISITES
    pip install azure-identity
    pip install azure-mgmt-authorization
# USAGE
    python role_assignments_create_for_subscription.py

    Before run the sample, please set the values of the client ID, tenant ID and client secret
    of the AAD application as environment variables: AZURE_CLIENT_ID, AZURE_TENANT_ID,
    AZURE_CLIENT_SECRET. For more info about how to get the value, please see:
    https://docs.microsoft.com/azure/active-directory/develop/howto-create-service-principal-portal
"""


def main():
    client = AuthorizationManagementClient(
        credential=DefaultAzureCredential(),
        subscription_id="SUBSCRIPTION_ID",
    )

    response = client.role_assignments.create(
        scope="subscriptions/a925f2f7-5c63-4b7b-8799-25a5f97bc3b2",
        role_assignment_name="05c5a614-a7d6-4502-b150-c2fb455033ff",
        parameters={
            "properties": {
                "principalId": "ce2ce14e-85d7-4629-bdbc-454d0519d987",
                "principalType": "User",
                "roleDefinitionId": "/subscriptions/a925f2f7-5c63-4b7b-8799-25a5f97bc3b2/providers/Microsoft.Authorization/roleDefinitions/0b5fe924-9a61-425c-96af-cfe6e287ca2d",
            }
        },
    )
    print(response)


# x-ms-original-file: 2022-04-01/RoleAssignments_CreateForSubscription.json
if __name__ == "__main__":
    main()



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

import pytest
from devtools_testutils import test_proxy, remove_batch_sanitizers
from dotenv import load_dotenv, find_dotenv

if not load_dotenv(find_dotenv(filename="tests.env"), override=True):
    print("Failed to apply environment variables for azure-ai-inference client library tests.")


# autouse=True will trigger this fixture on each pytest run, even if it's not explicitly used by a test method
@pytest.fixture(scope="session", autouse=True)
def start_proxy(test_proxy):
    return


@pytest.fixture(scope="session", autouse=True)
def add_sanitizers(test_proxy):
    # Remove the following sanitizers since certain fields are needed in tests and are non-sensitive:
    #  - AZSDK3493: $..name
    remove_batch_sanitizers(["AZSDK3493"])



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
# pylint: disable=wrong-import-position

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ._patch import *  # pylint: disable=unused-wildcard-import


from ._models import (  # type: ignore
    ApiEntityReference,
    ApplicationGateway,
    ApplicationGatewayBackendAddressPool,
    ArmResource,
    AzureFileVolume,
    CachedImages,
    Capabilities,
    CapabilitiesCapabilities,
    CloudError,
    CloudErrorBody,
    ConfidentialComputeProperties,
    ConfigMap,
    Container,
    ContainerAttachResponse,
    ContainerExec,
    ContainerExecRequest,
    ContainerExecRequestTerminalSize,
    ContainerExecResponse,
    ContainerGroup,
    ContainerGroupDiagnostics,
    ContainerGroupIdentity,
    ContainerGroupProfile,
    ContainerGroupProfilePatch,
    ContainerGroupProfileProperties,
    ContainerGroupProfileReferenceDefinition,
    ContainerGroupProfileStub,
    ContainerGroupProperties,
    ContainerGroupPropertiesInstanceView,
    ContainerGroupSubnetId,
    ContainerHttpGet,
    ContainerPort,
    ContainerProbe,
    ContainerProperties,
    ContainerPropertiesInstanceView,
    ContainerState,
    DeploymentExtensionSpec,
    DeploymentExtensionSpecProperties,
    DnsConfiguration,
    ElasticProfile,
    ElasticProfileContainerGroupNamingPolicy,
    ElasticProfileContainerGroupNamingPolicyGuidNamingPolicy,
    EncryptionProperties,
    EnvironmentVariable,
    ErrorAdditionalInfo,
    ErrorDetail,
    ErrorResponse,
    Event,
    FileShare,
    FileShareProperties,
    GitRepoVolume,
    GpuResource,
    HttpHeader,
    IdentityAccessControl,
    IdentityAcls,
    ImageRegistryCredential,
    InitContainerDefinition,
    InitContainerPropertiesDefinition,
    InitContainerPropertiesDefinitionInstanceView,
    IpAddress,
    LoadBalancer,
    LoadBalancerBackendAddressPool,
    LogAnalytics,
    Logs,
    ManagedServiceIdentity,
    NGroup,
    NGroupCGPropertyContainer,
    NGroupCGPropertyContainerProperties,
    NGroupCGPropertyVolume,
    NGroupContainerGroupProperties,
    NGroupIdentity,
    NGroupPatch,
    NGroupProperties,
    NetworkProfile,
    Operation,
    OperationDisplay,
    PlacementProfile,
    Port,
    ProxyResource,
    Resource,
    ResourceLimits,
    ResourceRequests,
    ResourceRequirements,
    SandboxGroup,
    SandboxGroupAccessToken,
    SandboxGroupNetworkProfile,
    SandboxGroupProperties,
    SandboxGroupTagsUpdate,
    SecretReference,
    SecurityContextCapabilitiesDefinition,
    SecurityContextDefinition,
    StandbyPoolProfileDefinition,
    StorageProfile,
    SubnetReference,
    SystemData,
    TrackedResource,
    UpdateProfile,
    UpdateProfileRollingUpdateProfile,
    Usage,
    UsageName,
    UserAssignedIdentities,
    UserAssignedIdentity,
    Volume,
    VolumeMount,
)

from ._enums import (  # type: ignore
    AzureFileShareAccessTier,
    AzureFileShareAccessType,
    ContainerGroupIpAddressType,
    ContainerGroupNetworkProtocol,
    ContainerGroupPriority,
    ContainerGroupRestartPolicy,
    ContainerGroupSku,
    ContainerInstanceOperationsOrigin,
    ContainerNetworkProtocol,
    CreatedByType,
    DnsNameLabelReusePolicy,
    GpuSku,
    IdentityAccessLevel,
    LogAnalyticsLogType,
    ManagedServiceIdentityType,
    NGroupProvisioningState,
    NGroupUpdateMode,
    OperatingSystemTypes,
    ResourceIdentityType,
    SandboxGroupProvisioningState,
    Scheme,
)
from ._patch import __all__ as _patch_all
from ._patch import *
from ._patch import patch_sdk as _patch_sdk

__all__ = [
    "ApiEntityReference",
    "ApplicationGateway",
    "ApplicationGatewayBackendAddressPool",
    "ArmResource",
    "AzureFileVolume",
    "CachedImages",
    "Capabilities",
    "CapabilitiesCapabilities",
    "CloudError",
    "CloudErrorBody",
    "ConfidentialComputeProperties",
    "ConfigMap",
    "Container",
    "ContainerAttachResponse",
    "ContainerExec",
    "ContainerExecRequest",
    "ContainerExecRequestTerminalSize",
    "ContainerExecResponse",
    "ContainerGroup",
    "ContainerGroupDiagnostics",
    "ContainerGroupIdentity",
    "ContainerGroupProfile",
    "ContainerGroupProfilePatch",
    "ContainerGroupProfileProperties",
    "ContainerGroupProfileReferenceDefinition",
    "ContainerGroupProfileStub",
    "ContainerGroupProperties",
    "ContainerGroupPropertiesInstanceView",
    "ContainerGroupSubnetId",
    "ContainerHttpGet",
    "ContainerPort",
    "ContainerProbe",
    "ContainerProperties",
    "ContainerPropertiesInstanceView",
    "ContainerState",
    "DeploymentExtensionSpec",
    "DeploymentExtensionSpecProperties",
    "DnsConfiguration",
    "ElasticProfile",
    "ElasticProfileContainerGroupNamingPolicy",
    "ElasticProfileContainerGroupNamingPolicyGuidNamingPolicy",
    "EncryptionProperties",
    "EnvironmentVariable",
    "ErrorAdditionalInfo",
    "ErrorDetail",
    "ErrorResponse",
    "Event",
    "FileShare",
    "FileShareProperties",
    "GitRepoVolume",
    "GpuResource",
    "HttpHeader",
    "IdentityAccessControl",
    "IdentityAcls",
    "ImageRegistryCredential",
    "InitContainerDefinition",
    "InitContainerPropertiesDefinition",
    "InitContainerPropertiesDefinitionInstanceView",
    "IpAddress",
    "LoadBalancer",
    "LoadBalancerBackendAddressPool",
    "LogAnalytics",
    "Logs",
    "ManagedServiceIdentity",
    "NGroup",
    "NGroupCGPropertyContainer",
    "NGroupCGPropertyContainerProperties",
    "NGroupCGPropertyVolume",
    "NGroupContainerGroupProperties",
    "NGroupIdentity",
    "NGroupPatch",
    "NGroupProperties",
    "NetworkProfile",
    "Operation",
    "OperationDisplay",
    "PlacementProfile",
    "Port",
    "ProxyResource",
    "Resource",
    "ResourceLimits",
    "ResourceRequests",
    "ResourceRequirements",
    "SandboxGroup",
    "SandboxGroupAccessToken",
    "SandboxGroupNetworkProfile",
    "SandboxGroupProperties",
    "SandboxGroupTagsUpdate",
    "SecretReference",
    "SecurityContextCapabilitiesDefinition",
    "SecurityContextDefinition",
    "StandbyPoolProfileDefinition",
    "StorageProfile",
    "SubnetReference",
    "SystemData",
    "TrackedResource",
    "UpdateProfile",
    "UpdateProfileRollingUpdateProfile",
    "Usage",
    "UsageName",
    "UserAssignedIdentities",
    "UserAssignedIdentity",
    "Volume",
    "VolumeMount",
    "AzureFileShareAccessTier",
    "AzureFileShareAccessType",
    "ContainerGroupIpAddressType",
    "ContainerGroupNetworkProtocol",
    "ContainerGroupPriority",
    "ContainerGroupRestartPolicy",
    "ContainerGroupSku",
    "ContainerInstanceOperationsOrigin",
    "ContainerNetworkProtocol",
    "CreatedByType",
    "DnsNameLabelReusePolicy",
    "GpuSku",
    "IdentityAccessLevel",
    "LogAnalyticsLogType",
    "ManagedServiceIdentityType",
    "NGroupProvisioningState",
    "NGroupUpdateMode",
    "OperatingSystemTypes",
    "ResourceIdentityType",
    "SandboxGroupProvisioningState",
    "Scheme",
]
__all__.extend([p for p in _patch_all if p not in __all__])  # pyright: ignore
_patch_sdk()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
"""Protocol conformance tests for handler-driven persistence (US1).

Verifies FR-001 (no persistence before handler runs),
FR-002 (bg=true: Create at response.created, Update at terminal),
FR-003 (bg=false: single Create at terminal state).

Python port of HandlerDrivenPersistenceTests.

NOTE: The reference tests use a RecordingProvider (spy) to verify exactly when
CreateResponseAsync and UpdateResponseAsync are called. The Python SDK uses
the in-memory FoundryStorageProvider (default) which we probe via GET to
confirm persistence timing.
"""

from __future__ import annotations

import asyncio
import json as _json
from typing import Any

import pytest

from azure.ai.agentserver.responses import ResponsesAgentServerHost
from azure.ai.agentserver.responses._id_generator import IdGenerator
from azure.ai.agentserver.responses.streaming._event_stream import ResponseEventStream

# ════════════════════════════════════════════════════════════
# Async ASGI client
# ════════════════════════════════════════════════════════════


class _AsgiResponse:
    def __init__(self, status_code: int, body: bytes, headers: list[tuple[bytes, bytes]]) -> None:
        self.status_code = status_code
        self.body = body
        self.headers = headers

    def json(self) -> Any:
        return _json.loads(self.body)


class _AsyncAsgiClient:
    def __init__(self, app: Any) -> None:
        self._app = app

    @staticmethod
    def _build_scope(method: str, path: str, body: bytes) -> dict[str, Any]:
        headers: list[tuple[bytes, bytes]] = []
        query_string = b""
        if "?" in path:
            path, qs = path.split("?", 1)
            query_string = qs.encode()
        if body:
            headers = [
                (b"content-type", b"application/json"),
                (b"content-length", str(len(body)).encode()),
            ]
        return {
            "type": "http",
            "asgi": {"version": "3.0"},
            "http_version": "1.1",
            "method": method,
            "headers": headers,
            "scheme": "http",
            "path": path,
            "raw_path": path.encode(),
            "query_string": query_string,
            "server": ("localhost", 80),
            "client": ("127.0.0.1", 123),
            "root_path": "",
        }

    async def request(self, method: str, path: str, *, json_body: dict[str, Any] | None = None) -> _AsgiResponse:
        body = _json.dumps(json_body).encode() if json_body else b""
        scope = self._build_scope(method, path, body)
        status_code: int | None = None
        response_headers: list[tuple[bytes, bytes]] = []
        body_parts: list[bytes] = []
        request_sent = False
        response_done = asyncio.Event()

        async def receive() -> dict[str, Any]:
            nonlocal request_sent
            if not request_sent:
                request_sent = True
                return {"type": "http.request", "body": body, "more_body": False}
            await response_done.wait()
            return {"type": "http.disconnect"}

        async def send(message: dict[str, Any]) -> None:
            nonlocal status_code, response_headers
            if message["type"] == "http.response.start":
                status_code = message["status"]
                response_headers = message.get("headers", [])
            elif message["type"] == "http.response.body":
                chunk = message.get("body", b"")
                if chunk:
                    body_parts.append(chunk)
                if not message.get("more_body", False):
                    response_done.set()

        await self._app(scope, receive, send)
        assert status_code is not None
        return _AsgiResponse(status_code=status_code, body=b"".join(body_parts), headers=response_headers)

    async def get(self, path: str) -> _AsgiResponse:
        return await self.request("GET", path)

    async def post(self, path: str, *, json_body: dict[str, Any] | None = None) -> _AsgiResponse:
        return await self.request("POST", path, json_body=json_body)


# ════════════════════════════════════════════════════════════
# Helpers
# ════════════════════════════════════════════════════════════


def _build_client(handler: Any) -> _AsyncAsgiClient:
    app = ResponsesAgentServerHost()
    app.response_handler(handler)
    return _AsyncAsgiClient(app)


async def _ensure_task_done(task: asyncio.Task[Any], handler: Any, timeout: float = 5.0) -> None:
    for attr in vars(handler):
        obj = getattr(handler, attr, None)
        if isinstance(obj, asyncio.Event):
            obj.set()
    if not task.done():
        try:
            await asyncio.wait_for(task, timeout=timeout)
        except (asyncio.TimeoutError, Exception):
            task.cancel()
            try:
                await task
            except (asyncio.CancelledError, Exception):
                pass


async def _wait_for_background_completion(client: _AsyncAsgiClient, response_id: str, timeout: float = 5.0) -> None:
    for _ in range(int(timeout / 0.05)):
        resp = await client.get(f"/responses/{response_id}")
        if resp.status_code == 404:
            await asyncio.sleep(0.05)
            continue
        doc = resp.json()
        if doc.get("status") in ("completed", "failed", "incomplete", "cancelled"):
            return
        await asyncio.sleep(0.05)
    raise TimeoutError(f"Response {response_id} did not reach terminal state within {timeout}s")


def _make_delaying_handler():
    """Handler that signals when started, then waits for a gate before yielding any events.

    Used to test FR-001: no persistence before handler runs.
    """
    started = asyncio.Event()
    gate = asyncio.Event()

    def handler(request: Any, context: Any, cancellation_signal: Any):
        async def _events():
            started.set()
            await gate.wait()
            stream = ResponseEventStream(
                response_id=context.response_id,
                model=getattr(request, "model", None),
            )
            yield stream.emit_created()
            yield stream.emit_completed()

        return _events()

    handler.started = started
    handler.gate = gate
    return handler


def _make_simple_handler():
    """Handler that emits created + completed immediately."""

    def handler(request: Any, context: Any, cancellation_signal: Any):
        async def _events():
            stream = ResponseEventStream(
                response_id=context.response_id,
                model=getattr(request, "model", None),
            )
            yield stream.emit_created()
            yield stream.emit_completed()

        return _events()

    return handler


# ════════════════════════════════════════════════════════════
# T015: bg+stream — provider NOT called until response.created
#
# FR-001: No persistence before handler emits response.created.
# Verifies that GET returns 404 before response.created is emitted.
# ════════════════════════════════════════════════════════════


@pytest.mark.asyncio
async def test_bg_stream_not_persisted_until_response_created() -> None:
    """T015/FR-001 — bg+stream: response not accessible before response.created."""
    handler = _make_delaying_handler()
    client = _build_client(handler)
    response_id = IdGenerator.new_response_id()

    post_task = asyncio.create_task(
        client.post(
            "/responses",
            json_body={
                "response_id": response_id,
                "model": "test",
                "background": True,
                "stream": True,
            },
        )
    )
    try:
        await asyncio.wait_for(handler.started.wait(), timeout=5.0)
        await asyncio.sleep(0.1)

        # GET before response.created — should NOT be accessible yet
        get_resp = await client.get(f"/responses/{response_id}")
        assert get_resp.status_code == 404, (
            f"FR-001: response should not be persisted before response.created, got status {get_resp.status_code}"
        )

        # Release handler → response.created will be yielded
        handler.gate.set()
        await asyncio.wait_for(post_task, timeout=5.0)
    finally:
        await _ensure_task_done(post_task, handler)

    # After handler completes, response should be accessible
    await _wait_for_background_completion(client, response_id)
    get_after = await client.get(f"/responses/{response_id}")
    assert get_after.status_code == 200


# ════════════════════════════════════════════════════════════
# T016: bg+nostream — provider NOT called until response.created
# ════════════════════════════════════════════════════════════


@pytest.mark.asyncio
async def test_bg_nostream_not_persisted_until_response_created() -> None:
    """T016/FR-001 — bg+nostream: response not accessible before response.created."""
    handler = _make_delaying_handler()
    client = _build_client(handler)
    response_id = IdGenerator.new_response_id()

    post_task = asyncio.create_task(
        client.post(
            "/responses",
            json_body={
                "response_id": response_id,
                "model": "test",
                "background": True,
            },
        )
    )
    try:
        await asyncio.wait_for(handler.started.wait(), timeout=5.0)
        await asyncio.sleep(0.1)

        # GET before response.created — should NOT be accessible
        get_resp = await client.get(f"/responses/{response_id}")
        assert get_resp.status_code == 404, (
            f"FR-001: response should not be persisted before response.created, got status {get_resp.status_code}"
        )

        # Release handler
        handler.gate.set()
        await asyncio.wait_for(post_task, timeout=5.0)
    finally:
        await _ensure_task_done(post_task, handler)

    # After handler completes, response should be accessible
    await _wait_for_background_completion(client, response_id)
    get_after = await client.get(f"/responses/{response_id}")
    assert get_after.status_code == 200


# ════════════════════════════════════════════════════════════
# T017: bg=true — exactly 1 Create + 1 Update
#
# FR-002: bg mode persists Create at response.created, Update at terminal.
# We verify via GET that the response is accessible during in-progress
# and that after completion the status is updated.
# ════════════════════════════════════════════════════════════


@pytest.mark.asyncio
async def test_bg_mode_response_accessible_during_and_after_handler() -> None:
    """T017/FR-002 — bg mode: response accessible at in_progress and completed."""
    started = asyncio.Event()
    release = asyncio.Event()

    def handler(request: Any, context: Any, cancellation_signal: Any):
        async def _events():
            stream = ResponseEventStream(
                response_id=context.response_id,
                model=getattr(request, "model", None),
            )
            yield stream.emit_created()
            yield stream.emit_in_progress()
            started.set()
            while not release.is_set():
                if cancellation_signal.is_set():
                    return
                await asyncio.sleep(0.01)
            yield stream.emit_completed()

        return _events()

    handler.started = started
    handler.release = release

    client = _build_client(handler)
    response_id = IdGenerator.new_response_id()

    post_task = asyncio.create_task(
        client.post(
            "/responses",
            json_body={
                "response_id": response_id,
                "model": "test",
                "background": True,
            },
        )
    )
    try:
        await asyncio.wait_for(started.wait(), timeout=5.0)

        # During handler execution — response should be accessible as in_progress
        get_mid = await client.get(f"/responses/{response_id}")
        assert get_mid.status_code == 200
        assert get_mid.json()["status"] == "in_progress"

        # Release handler
        release.set()
        await asyncio.wait_for(post_task, timeout=5.0)
    finally:
        for attr in ("started", "release"):
            obj = getattr(handler, attr, None)
            if isinstance(obj, asyncio.Event):
                obj.set()
        if not post_task.done():
            post_task.cancel()
            try:
                await post_task
            except (asyncio.CancelledError, Exception):
                pass

    # After completion — response updated to completed
    await _wait_for_background_completion(client, response_id)
    get_final = await client.get(f"/responses/{response_id}")
    assert get_final.status_code == 200
    assert get_final.json()["status"] == "completed"


# ════════════════════════════════════════════════════════════
# T018: bg=false — single Create at terminal (no mid-flight GET)
#
# FR-003: non-bg mode does a single Create at terminal. Not accessible mid-flight.
# ════════════════════════════════════════════════════════════


@pytest.mark.asyncio
async def test_non_bg_not_accessible_until_terminal() -> None:
    """T018/FR-003 — non-bg: response only accessible after terminal state."""
    started = asyncio.Event()
    release = asyncio.Event()

    def handler(request: Any, context: Any, cancellation_signal: Any):
        async def _events():
            stream = ResponseEventStream(
                response_id=context.response_id,
                model=getattr(request, "model", None),
            )
            yield stream.emit_created()
            yield stream.emit_in_progress()
            started.set()
            await release.wait()
            yield stream.emit_completed()

        return _events()

    handler.started = started
    handler.release = release

    client = _build_client(handler)
    response_id = IdGenerator.new_response_id()

    post_task = asyncio.create_task(
        client.post(
            "/responses",
            json_body={
                "response_id": response_id,
                "model": "test",
                "stream": False,
                "background": False,
            },
        )
    )
    try:
        await asyncio.wait_for(started.wait(), timeout=5.0)

        # During non-bg handler execution — response should NOT be accessible
        get_mid = await client.get(f"/responses/{response_id}")
        assert get_mid.status_code == 404, (
            f"FR-003: non-bg response should not be accessible mid-flight, got {get_mid.status_code}"
        )

        release.set()
        post_resp = await asyncio.wait_for(post_task, timeout=5.0)
        assert post_resp.status_code == 200
    finally:
        started.set()
        release.set()
        if not post_task.done():
            post_task.cancel()
            try:
                await post_task
            except (asyncio.CancelledError, Exception):
                pass

    # After terminal — response is accessible as completed
    get_after = await client.get(f"/responses/{response_id}")
    assert get_after.status_code == 200
    assert get_after.json()["status"] == "completed"



# pylint: disable=line-too-long,useless-suppression
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from azure.identity import DefaultAzureCredential

from azure.mgmt.datamigration import DataMigrationManagementClient

"""
# PREREQUISITES
    pip install azure-identity
    pip install azure-mgmt-datamigration
# USAGE
    python services_check_children_name_availability.py

    Before run the sample, please set the values of the client ID, tenant ID and client secret
    of the AAD application as environment variables: AZURE_CLIENT_ID, AZURE_TENANT_ID,
    AZURE_CLIENT_SECRET. For more info about how to get the value, please see:
    https://docs.microsoft.com/azure/active-directory/develop/howto-create-service-principal-portal
"""


def main():
    client = DataMigrationManagementClient(
        credential=DefaultAzureCredential(),
        subscription_id="fc04246f-04c5-437e-ac5e-206a19e7193f",
    )

    response = client.services.check_children_name_availability(
        group_name="DmsSdkRg",
        service_name="DmsSdkService",
        parameters={"name": "Task1", "type": "tasks"},
    )
    print(response)


# x-ms-original-file: specification/datamigration/resource-manager/Microsoft.DataMigration/stable/2025-06-30/examples/Services_CheckChildrenNameAvailability.json
if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from azure.core.pipeline.policies import HTTPPolicy
from azure.core.pipeline import PipelineRequest
from . import token_utils


class EntraTokenGuardPolicy(HTTPPolicy):
    """A pipeline policy that caches the response for a given Entra token and reuses it if valid."""

    def __init__(self):
        super().__init__()
        self._entra_token_cache = None
        self._response_cache = None

    def send(self, request: PipelineRequest):
        cache_valid, token = token_utils.is_entra_token_cache_valid(self._entra_token_cache, request)
        if cache_valid and token_utils.is_acs_token_cache_valid(self._response_cache):
            response = self._response_cache
        else:
            self._entra_token_cache = token
            response = self.next.send(request)
            self._response_cache = response
        if response is None:
            raise RuntimeError("Failed to obtain a valid PipelineResponse in EntraTokenGuardPolicy.send")
        return response



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""Observability setup (logging + tracing) for AgentServerHost.

This module is the single source of truth for all logging handlers,
tracing exporters, and span operations:

**Logging & tracing setup:**

- :func:`configure_observability` — default one-time setup called by
  ``AgentServerHost.__init__``.  Configures:

  - Console ``StreamHandler`` on the **root** logger (so both SDK and
    user ``logging.info()`` calls are visible).
  - Suppression of noisy Azure Core HTTP logging policy output.
  - Trace and log export via ``microsoft-opentelemetry`` distro (auto-detects
    Azure Monitor from ``APPLICATIONINSIGHTS_CONNECTION_STRING`` and OTLP
    from ``OTEL_EXPORTER_OTLP_ENDPOINT``).

  Users may pass a custom callable (or ``None``) via the
  ``configure_observability`` constructor parameter to override or
  disable this default setup.

**Span operations:**

- :class:`TraceContextMiddleware` — ASGI middleware that extracts W3C trace
  context and baggage from incoming headers
- :func:`end_span` / :func:`record_error` — span lifecycle helpers
- :func:`trace_stream` — wrap streaming responses with span lifecycle
- :func:`set_current_span` / :func:`detach_context` — explicit context management

OpenTelemetry is a required dependency — these functions always create
real spans.  Azure Monitor export is optional (auto-configured by the distro).
"""
import logging
import os
from collections.abc import AsyncIterable, AsyncIterator  # pylint: disable=import-error
from typing import Any, Optional, Union

from opentelemetry import baggage as _otel_baggage, context as _otel_context, trace

from . import _config

_Content = Union[str, bytes, memoryview]

# GenAI semantic convention attribute keys
_ATTR_SERVICE_NAME = "service.name"
_ATTR_GEN_AI_SYSTEM = "gen_ai.system"
_ATTR_GEN_AI_PROVIDER_NAME = "gen_ai.provider.name"
_ATTR_GEN_AI_AGENT_ID = "gen_ai.agent.id"
_ATTR_GEN_AI_AGENT_BLUEPRINT_ID = "microsoft.a365.agent.blueprint.id"
_ATTR_GEN_AI_AGENT_TENANT_ID = "microsoft.tenant.id"
_ATTR_GEN_AI_AGENT_NAME = "gen_ai.agent.name"
_ATTR_GEN_AI_AGENT_VERSION = "gen_ai.agent.version"
_ATTR_GEN_AI_RESPONSE_ID = "gen_ai.response.id"
_ATTR_GEN_AI_OPERATION_NAME = "gen_ai.operation.name"
_ATTR_GEN_AI_CONVERSATION_ID = "gen_ai.conversation.id"
_ATTR_FOUNDRY_PROJECT_ID = "microsoft.foundry.project.id"
_ATTR_SESSION_ID = "microsoft.session.id"

# Baggage keys consumed by tracing.
# Currently, the invocations package sets the session-id baggage key.
# The conversation-id baggage key is defined for propagation/mapping, but
# is not currently set elsewhere in this repo.  Incoming requests from
# the calling service may carry either key as W3C baggage.
_BAGGAGE_SESSION_ID = "azure.ai.agentserver.session_id"
_BAGGAGE_CONVERSATION_ID = "azure.ai.agentserver.conversation_id"
_BAGGAGE_INVOCATION_ID = "azure.ai.agentserver.invocation_id"

_ATTR_INVOCATION_ID = "azure.ai.agentserver.invocations.invocation_id"

_SERVICE_NAME_VALUE = "azure.ai.agentserver"
_GEN_AI_SYSTEM_VALUE = "azure.ai.agentserver"
_GEN_AI_PROVIDER_NAME_VALUE = "AzureAI Hosted Agents"

logger = logging.getLogger("azure.ai.agentserver")


# ======================================================================
# Public API: observability setup
# ======================================================================

# Sentinel attribute name set on the console handler to prevent adding
# duplicates across multiple AgentServerHost instantiations.
_CONSOLE_HANDLER_ATTR = "_agentserver_console"


def configure_observability(
    *,
    connection_string: Optional[str] = None,
    log_level: Optional[str] = None,
    enable_sensitive_data: bool = False,
) -> None:
    """Default observability setup: console logging + tracing/OTel export.

    Attaches a formatted ``StreamHandler`` to the **root** logger so that
    both SDK and user ``logging.info()`` calls are visible on the console.
    Then configures OpenTelemetry tracing and log export (Azure Monitor
    and/or OTLP) when a connection string or OTLP endpoint is available.

    Pass this function (or a custom replacement) as the
    ``configure_observability`` parameter of :class:`AgentServerHost`.
    Pass ``None`` to disable all SDK-managed observability setup.

    :keyword connection_string: Application Insights connection string.
    :paramtype connection_string: str or None
    :keyword log_level: Log level name (e.g. ``"INFO"``, ``"DEBUG"``).
    :paramtype log_level: str or None
    :keyword enable_sensitive_data: Enable sensitive data recording
        (prompts, tool arguments, results) for Agent Framework SDK
        instrumentation. Defaults to False.
    :paramtype enable_sensitive_data: bool
    """
    # Console logging on the root logger so user logs are also visible.
    resolved_level = _config.resolve_log_level(log_level)
    root = logging.getLogger()
    root.setLevel(resolved_level)
    # Only add a console handler if root doesn't already have one.
    # Check for our sentinel-marked handler AND any existing StreamHandler
    # (e.g. from user's logging.basicConfig() or framework setup) to
    # prevent duplicate output on stderr.
    _has_console = any(
        getattr(h, _CONSOLE_HANDLER_ATTR, False)
        or (isinstance(h, logging.StreamHandler) and not isinstance(h, logging.FileHandler))
        for h in root.handlers
    )
    if not _has_console:
        _console = logging.StreamHandler()
        _console.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(name)s: %(message)s"))
        setattr(_console, _CONSOLE_HANDLER_ATTR, True)
        root.addHandler(_console)

    # Suppress the noisy Azure Core HTTP logging policy logger.
    logging.getLogger("azure.core.pipeline.policies.http_logging_policy").setLevel(logging.WARNING)

    # Tracing and OTel export
    _configure_tracing(connection_string=connection_string, enable_sensitive_data=enable_sensitive_data)


def _configure_tracing(connection_string: Optional[str] = None, enable_sensitive_data: bool = False) -> None:
    """Configure OpenTelemetry exporters via the microsoft-opentelemetry distro.

    Internal helper called by :func:`configure_observability`.

    :param connection_string: Application Insights connection string.
        When provided, traces and logs are exported to Azure Monitor.
    :type connection_string: str or None
    :param enable_sensitive_data: Enable sensitive data recording for
        Agent Framework SDK instrumentation.
    :type enable_sensitive_data: bool
    """
    resource = _create_resource()
    if resource is None:
        logger.warning("Failed to create OTel resource — tracing will not be configured.")
        return

    # Build custom processors
    agent_name = _config.resolve_agent_name() or None
    agent_version = _config.resolve_agent_version() or None
    project_id = _config.resolve_project_id() or None
    agent_id = _config.resolve_agent_id() or None
    agent_blueprint_id = _config.resolve_agent_blueprint_id() or None
    agent_tenant_id = _config.resolve_agent_tenant_id() or None

    span_processors = [
        _FoundryEnrichmentSpanProcessor(
            agent_name=agent_name, agent_version=agent_version,
            agent_id=agent_id, project_id=project_id,
            agent_blueprint_id=agent_blueprint_id,
            agent_tenant_id=agent_tenant_id,
        ),
    ]
    log_record_processors = [_BaggageLogRecordProcessor()]  # type: ignore[list-item]

    try:
        _setup_distro_export(
            resource=resource,
            span_processors=span_processors,
            log_record_processors=log_record_processors,
            connection_string=connection_string,
            enable_sensitive_data=enable_sensitive_data,
        )
        logger.info("Tracing configured successfully via microsoft-opentelemetry distro.")
    except ImportError:
        logger.warning("microsoft-opentelemetry is not installed — tracing export disabled.")
        # Still set up TracerProvider with enrichment processor so spans are created
        _ensure_trace_provider(resource, span_processors)


def _setup_distro_export(
    *,
    resource: Any,
    span_processors: list[Any],
    log_record_processors: list[Any],
    connection_string: Optional[str] = None,
    enable_sensitive_data: bool = False,
) -> None:
    """Delegate to microsoft-opentelemetry distro for exporter configuration.

    Separated into its own function so tests can easily mock it without
    intercepting lazy imports.

    :keyword resource: OTel resource describing this service.
    :keyword span_processors: Span processors to register.
    :keyword log_record_processors: Log record processors to register.
    :keyword connection_string: Application Insights connection string.
    :keyword enable_sensitive_data: Enable sensitive data recording for
        Agent Framework SDK instrumentation.
    """
    from microsoft.opentelemetry import use_microsoft_opentelemetry

    kwargs: dict[str, Any] = {
        "resource": resource,
        "span_processors": span_processors,
        "log_record_processors": log_record_processors,
        "enable_sensitive_data": enable_sensitive_data,
    }

    # Azure Monitor export is off by default in the distro — enable it
    # when a connection string is available.
    if connection_string:
        kwargs["enable_azure_monitor"] = True
        kwargs["azure_monitor_connection_string"] = connection_string

    # A365 tracing export — enabled only in hosted environments.
    if (
        os.environ.get("FOUNDRY_HOSTING_ENVIRONMENT", "")
        and os.environ.get("FOUNDRY_AGENT365_TRACING_ENABLED", "").lower() in ("true", "1")
    ):
        kwargs["enable_a365"] = True
        kwargs["a365_use_s2s_endpoint"] = True
        kwargs["a365_enable_observability_exporter"] = True
        kwargs["a365_observability_scope_override"] = "api://9b975845-388f-4429-889e-eab1ef63949c/.default"

    use_microsoft_opentelemetry(**kwargs)


# ======================================================================
# Public API: span operations
# ======================================================================


class TraceContextMiddleware:
    """Pure-ASGI middleware that propagates W3C trace context and baggage.

    Extracts ``traceparent``, ``tracestate``, and ``baggage`` headers from
    incoming HTTP requests using the standard W3C propagators and attaches
    the resulting context for the duration of the request.  This ensures
    that any spans created downstream (e.g. by agent-framework / MAF) are
    automatically children of the caller's trace.

    This middleware does **not** create its own span — it only propagates
    the incoming context so that downstream instrumentation inherits it.

    :param app: The inner ASGI application.
    :type app: ASGIApp
    """

    def __init__(self, app: Any) -> None:
        self.app = app

    async def __call__(self, scope: Any, receive: Any, send: Any) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        # Build a simple dict of headers for the propagators
        raw_headers: list[tuple[bytes, bytes]] = scope.get("headers", [])
        headers = {
            k.decode("latin-1"): v.decode("latin-1")
            for k, v in raw_headers
        }

        # Use the global propagator to extract trace context + baggage
        from opentelemetry.propagate import extract  # pylint: disable=import-outside-toplevel
        ctx = extract(carrier=headers)

        # Add x-request-id as baggage for downstream propagation
        x_request_id = headers.get("x-request-id")
        if x_request_id:
            ctx = _otel_baggage.set_baggage(
                "x_request_id", x_request_id, context=ctx,
            )

        token = _otel_context.attach(ctx)
        try:
            await self.app(scope, receive, send)
        finally:
            try:
                _otel_context.detach(token)
            except ValueError:
                pass


def end_span(span: Any, exc: Optional[BaseException] = None) -> None:
    """End a span, optionally recording an error first.

    No-op when *span* is ``None``.

    :param span: The OTel span to end, or ``None``.
    :type span: any
    :param exc: Optional exception to record before ending.
    :type exc: BaseException or None
    """
    if span is None:
        return
    if exc is not None:
        record_error(span, exc)
    span.end()


def flush_spans(timeout_millis: int = 5000) -> None:
    """Flush all pending spans from the TracerProvider.

    The ``BatchSpanProcessor`` buffers spans and exports them on a timer
    (default 5 seconds).  In hosted sandbox environments the platform may
    suspend the process immediately after an HTTP response is sent, before
    the batch timer fires.  Calling this function after ending the request
    span ensures that all child spans — including short-lived ones created
    by third-party tracers (e.g. LangChain/LangGraph per-node spans) — are
    exported before the sandbox is frozen.

    No-op when the OTel SDK is not installed or the provider does not
    support ``force_flush``.

    :param timeout_millis: Maximum time to wait for the flush, in
        milliseconds.  Defaults to 5000 (5 seconds).
    :type timeout_millis: int
    """
    provider = trace.get_tracer_provider()
    flush = getattr(provider, "force_flush", None)
    if flush is not None:
        try:
            flush(timeout_millis)
        except Exception:  # pylint: disable=broad-exception-caught
            logger.debug("TracerProvider.force_flush() failed", exc_info=True)


def record_error(span: Any, exc: BaseException) -> None:
    """Record an exception and ERROR status on a span.

    Sets ``error.type`` and ``otel.status.description`` per OTel
    semantic conventions.

    :param span: The OTel span, or ``None``.
    :type span: any
    :param exc: The exception to record.
    :type exc: BaseException
    """
    if span is not None:
        span.set_status(trace.StatusCode.ERROR, str(exc))
        span.set_attribute("error.type", type(exc).__name__)
        span.record_exception(exc)


def set_current_span(span: Any) -> Any:
    """Set a span as the current span in the OTel context.

    This makes *span* the active parent for any child spans created
    by downstream code (e.g. framework handlers).  Without this,
    spans created inside the handler would become siblings rather
    than children of *span*.

    Returns a context token that **must** be passed to
    :func:`detach_context` when the scope ends.  No-op when *span*
    is ``None``.

    :param span: The OTel span to make current, or *None*.
    :type span: Any
    :return: A context token, or *None*.
    :rtype: Any
    """
    if span is None:
        return None
    ctx = trace.set_span_in_context(span)
    return _otel_context.attach(ctx)


def detach_context(token: Any) -> None:
    """Detach a context previously attached by :func:`set_current_span`.

    Best-effort no-op when *token* is ``None`` or when the token is no
    longer the current OpenTelemetry context.

    :param token: The token returned by :func:`set_current_span`.
    :type token: Any
    """
    if token is not None:
        try:
            _otel_context.detach(token)
        except ValueError:
            logging.getLogger(__name__).debug(
                "Ignoring OpenTelemetry context detach for a non-current token.",
                exc_info=True,
            )


async def trace_stream(
    iterator: AsyncIterable[_Content], span: Any
) -> AsyncIterator[_Content]:
    """Wrap a streaming body so the span covers the full transmission.

    Yields chunks unchanged.  Ends the span when the iterator is
    exhausted or raises an exception.

    :param iterator: The async iterable to wrap.
    :type iterator: AsyncIterable[str or bytes or memoryview]
    :param span: The OTel span to end on completion, or ``None``.
    :type span: any
    :return: An async iterator yielding chunks unchanged.
    :rtype: AsyncIterator[str or bytes or memoryview]
    """
    error: Optional[BaseException] = None
    try:
        async for chunk in iterator:
            yield chunk
    except BaseException as exc:
        error = exc
        raise
    finally:
        end_span(span, exc=error)
        flush_spans()


# ======================================================================
# Foundry enrichment span processor
# ======================================================================


class _FoundryEnrichmentSpanProcessor:
    """Adds Foundry identity attributes to ALL spans."""

    def __init__(
        self,
        *,
        agent_name: Optional[str] = None,
        agent_version: Optional[str] = None,
        agent_id: Optional[str] = None,
        project_id: Optional[str] = None,
        agent_blueprint_id: Optional[str] = None,
        agent_tenant_id: Optional[str] = None,
    ) -> None:
        self.agent_name = agent_name
        self.agent_version = agent_version
        self.agent_id = agent_id
        self.project_id = project_id
        self.agent_blueprint_id = agent_blueprint_id
        self.agent_tenant_id = agent_tenant_id

    def on_start(self, span: Any, parent_context: Any = None) -> None:
        if self.project_id:
            span.set_attribute(_ATTR_FOUNDRY_PROJECT_ID, self.project_id)

        # Stamp session and conversation IDs from baggage so child spans
        # created by frameworks (LangChain, Semantic Kernel, etc.) inherit them.
        ctx = parent_context or _otel_context.get_current()
        session_id = _otel_baggage.get_baggage(_BAGGAGE_SESSION_ID, context=ctx)
        if session_id:
            span.set_attribute(_ATTR_SESSION_ID, session_id)
        conversation_id = _otel_baggage.get_baggage(_BAGGAGE_CONVERSATION_ID, context=ctx)
        if conversation_id:
            span.set_attribute(_ATTR_GEN_AI_CONVERSATION_ID, conversation_id)
        invocation_id = _otel_baggage.get_baggage(_BAGGAGE_INVOCATION_ID, context=ctx)
        if invocation_id:
            span.set_attribute(_ATTR_INVOCATION_ID, invocation_id)

    def _on_ending(self, span: Any) -> None:
        # Set agent identity attributes at span end so they cannot be
        # overwritten by underlying frameworks (e.g. LangChain, Semantic Kernel).
        #
        # Workaround: opentelemetry-sdk <=1.40.0 sets _end_time before calling
        # _on_ending, which causes set_attribute() to silently no-op despite the
        # spec requiring mutability during OnEnding.  We write to _attributes
        # directly until the SDK is fixed.  The try/except guards against future
        # SDK changes that may rename or remove the internal field.
        # TODO: switch to span.set_attribute() once the SDK honours the spec.
        attrs = getattr(span, "_attributes", None)
        if attrs is None:
            return
        try:
            if self.agent_name:
                attrs[_ATTR_GEN_AI_AGENT_NAME] = self.agent_name
            if self.agent_version:
                attrs[_ATTR_GEN_AI_AGENT_VERSION] = self.agent_version
            if self.agent_id:
                attrs[_ATTR_GEN_AI_AGENT_ID] = self.agent_id
            if self.agent_blueprint_id:
                attrs[_ATTR_GEN_AI_AGENT_BLUEPRINT_ID] = self.agent_blueprint_id
            if self.agent_tenant_id:
                attrs[_ATTR_GEN_AI_AGENT_TENANT_ID] = self.agent_tenant_id
        except Exception:  # pylint: disable=broad-exception-caught
            logger.debug("Failed to enrich span attributes in _on_ending", exc_info=True)

    def on_end(self, span: Any) -> None:  # pylint: disable=unused-argument
        pass

    def shutdown(self) -> None:
        pass

    def force_flush(self, timeout_millis: int = 30000) -> bool:  # pylint: disable=unused-argument
        return True


class _BaggageLogRecordProcessor:
    """OTel log record processor that copies W3C Baggage entries into log attributes.

    Per container-image-spec §6.1, all baggage key-value pairs from the
    current span context should appear as attributes on every log record
    for end-to-end correlation.
    """

    def on_emit(self, log_data: Any) -> None:  # pylint: disable=unused-argument
        """Copy baggage entries into the log record's attributes.

        :param log_data: The log data being emitted.
        :type log_data: any
        """
        try:
            ctx = _otel_context.get_current()
            entries = _otel_baggage.get_all(context=ctx)
            if entries and hasattr(log_data, 'log_record') and log_data.log_record:
                for key, value in entries.items():
                    log_data.log_record.attributes[key] = value  # type: ignore[index]
        except Exception:  # pylint: disable=broad-except
            pass

    def shutdown(self) -> None:
        pass

    def force_flush(self, timeout_millis: int = 30000) -> bool:  # pylint: disable=unused-argument
        return True


# ======================================================================
# Internal: resource, provider, exporters
# ======================================================================


def _create_resource() -> Any:
    try:
        from opentelemetry.sdk.resources import Resource
    except ImportError:
        logger.warning("OTel SDK not installed — tracing resource creation failed.")
        return None
    # service.name maps to cloud_RoleName in App Insights
    agent_name = os.environ.get(_config._ENV_FOUNDRY_AGENT_NAME, "")  # pylint: disable=protected-access
    service_name = agent_name or _SERVICE_NAME_VALUE
    return Resource.create({_ATTR_SERVICE_NAME: service_name})


def _ensure_trace_provider(resource: Any, span_processors: Optional[list[Any]] = None) -> Any:
    """Get or create a TracerProvider, optionally adding span processors.

    Used as a fallback when the microsoft-opentelemetry distro is not installed.

    :param resource: OTel resource describing this service.
    :type resource: ~typing.Any
    :param span_processors: Optional span processors to register.
    :type span_processors: list[~typing.Any] or None
    """
    if resource is None:
        return None
    try:
        from opentelemetry.sdk.trace import TracerProvider as SdkTracerProvider
    except ImportError:
        return None
    current = trace.get_tracer_provider()
    if hasattr(current, "add_span_processor"):
        provider = current
    else:
        provider = SdkTracerProvider(resource=resource)
        trace.set_tracer_provider(provider)
    if span_processors and not getattr(provider, "_agentserver_processors_added", False):
        for proc in span_processors:
            provider.add_span_processor(proc)
        provider._agentserver_processors_added = True  # type: ignore[attr-defined]  # pylint: disable=protected-access
    return provider



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from copy import copy
from flask import (
    Response,
    Blueprint,
    request,
)
from .helpers import assert_with_message

multipart_api = Blueprint("multipart_api", __name__)

multipart_header_start = "multipart/form-data; boundary="

# NOTE: the flask behavior is different for aiohttp and requests
# in requests, we see the file content through request.form
# in aiohttp, we see the file through request.files


@multipart_api.route("/basic", methods=["POST"])
def basic():
    assert_with_message("content type", multipart_header_start, request.content_type[: len(multipart_header_start)])
    if request.files:
        # aiohttp
        assert_with_message("num files", 1, len(request.files))
        assert_with_message("has file named fileContent", True, bool(request.files.get("fileContent")))
        file_content = request.files["fileContent"]
        assert_with_message("file content type", "application/octet-stream", file_content.content_type)
        assert_with_message("filename", "fileContent", file_content.filename)
        assert_with_message(
            "has content disposition header", True, bool(file_content.headers.get("Content-Disposition"))
        )
        assert_with_message(
            "content disposition",
            'form-data; name="fileContent"; filename="fileContent"',
            file_content.headers["Content-Disposition"],
        )
    elif request.form:
        # requests
        assert_with_message("fileContent", "<file content>", request.form["fileContent"])
    else:
        return Response(status=400)  # should be either of these
    return Response(status=200)


@multipart_api.route("/data-and-files", methods=["POST"])
def data_and_files():
    if request.files:
        # aiohttp
        assert_with_message("content type", multipart_header_start, request.content_type[: len(multipart_header_start)])
        assert_with_message("has file named fileContent", True, bool(request.files.get("fileContent")))
        assert_with_message("message", "Hello, world!", request.form["message"])
        file_content = request.files["fileContent"]
        assert_with_message("file content type", "application/octet-stream", file_content.content_type)
        assert_with_message("filename", "fileContent", file_content.filename)
        assert_with_message(
            "has content disposition header", True, bool(file_content.headers.get("Content-Disposition"))
        )
        assert_with_message(
            "content disposition",
            'form-data; name="fileContent"; filename="fileContent"',
            file_content.headers["Content-Disposition"],
        )
    elif request.form:
        # requests
        assert_with_message("content type", multipart_header_start, request.content_type[: len(multipart_header_start)])
        assert_with_message("message", "Hello, world!", request.form["message"])
        assert_with_message("message", "<file content>", request.form["fileContent"])
    else:
        return Response(status=400)
    return Response(status=200)


@multipart_api.route("/data-and-files-tuple", methods=["POST"])
def data_and_files_tuple():
    assert_with_message("content type", multipart_header_start, request.content_type[: len(multipart_header_start)])
    assert_with_message("message", ["abc"], request.form["message"])
    assert_with_message("message", ["<file content>"], request.form["fileContent"])
    return Response(status=200)


@multipart_api.route("/non-seekable-filelike", methods=["POST"])
def non_seekable_filelike():
    assert_with_message("content type", multipart_header_start, request.content_type[: len(multipart_header_start)])
    if request.files:
        # aiohttp
        len_files = len(request.files)
        assert_with_message("num files", 1, len_files)
        assert_with_message("num files", 1, len(request.files))
        assert_with_message("has file named file", True, bool(request.files.get("file")))
        file = request.files["file"]
        assert_with_message("file content type", "application/octet-stream", file.content_type)
        assert_with_message("filename", "file", file.filename)
        assert_with_message("has content disposition header", True, bool(file.headers.get("Content-Disposition")))
        assert_with_message(
            "content disposition",
            'form-data; name="file"; filename="file"',
            file.headers["Content-Disposition"],
        )
    elif request.form:
        # requests
        assert_with_message("num files", 1, len(request.form))
    else:
        return Response(status=400)
    return Response(status=200)


@multipart_api.route("/request", methods=["POST"])
def multipart_request():
    body_as_str = (
        "--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        "Content-Type: application/http\r\n"
        "Content-ID: 0\r\n"
        "\r\n"
        "HTTP/1.1 202 Accepted\r\n"
        "x-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        "x-version: 2018-11-09\r\n"
        "\r\n"
        "--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        "Content-Type: application/http\r\n"
        "Content-ID: 2\r\n"
        "\r\n"
        "HTTP/1.1 404 The specified blob does not exist.\r\n"
        "x-error-code: BlobNotFound\r\n"
        "x-request-id: 778fdc83-801e-0000-62ff-0334671e2852\r\n"
        "x-version: 2018-11-09\r\n"
        "Content-Length: 216\r\n"
        "Content-Type: application/xml\r\n"
        "\r\n"
        '<?xml version="1.0" encoding="utf-8"?>\r\n'
        "<Error><Code>BlobNotFound</Code><Message>The specified blob does not exist.\r\n"
        "RequestId:778fdc83-801e-0000-62ff-0334671e2852\r\n"
        "Time:2018-06-14T16:46:54.6040685Z</Message></Error>\r\n"
        "--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--"
    )
    return Response(
        body_as_str.encode("ascii"),
        content_type="multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed",
    )


@multipart_api.route("/tuple-input-multiple-same-name", methods=["POST"])
def tuple_input_multiple_same_name():
    assert_with_message("content type", multipart_header_start, request.content_type[: len(multipart_header_start)])

    files = request.files.getlist("file")
    assert_with_message("num files", 2, len(files))

    file1 = files[0]
    assert_with_message("file content type", "image/pdf", file1.content_type)
    assert_with_message("filename", "firstFileName", file1.filename)

    file2 = files[1]
    assert_with_message("file content type", "image/png", file2.content_type)
    assert_with_message("filename", "secondFileName", file2.filename)
    return Response(status=200)


@multipart_api.route("/tuple-input-multiple-same-name-with-tuple-file-value", methods=["POST"])
def test_input_multiple_same_name_with_tuple_file_value():
    assert_with_message("content type", multipart_header_start, request.content_type[: len(multipart_header_start)])

    images = request.files.getlist("images")
    assert_with_message("num images", 2, len(images))

    tuple_image = images[0]
    assert_with_message("image content type", "image/png", tuple_image.content_type)
    assert_with_message("filename", "foo.png", tuple_image.filename)

    single_image = images[1]
    assert_with_message("file content type", "application/octet-stream", single_image.content_type)
    assert_with_message("filename", "foo.png", single_image.filename)
    return Response(status=200)


@multipart_api.route("/data-and-file-input-same-name", methods=["POST"])
def data_and_file_input_same_name():
    assert_with_message("content type", multipart_header_start, request.content_type[: len(multipart_header_start)])

    # use call to this function to check files
    tuple_input_multiple_same_name()

    assert_with_message("data items num", 1, len(request.form.keys()))
    assert_with_message("message", "Hello, world!", request.form["message"])

    return Response(status=200)



# coding: utf-8

# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from datetime import datetime
import pytest
from devtools_testutils import recorded_by_proxy
from azure.core.exceptions import (
    HttpResponseError,
)
from _router_test_case import RouterRecordedTestCase

from _decorators import RouterPreparers

from azure.communication.jobrouter import (
    JobRouterClient,
    JobRouterAdministrationClient,
)
from azure.communication.jobrouter.models import (
    LongestIdleMode,
    RouterChannel,
    RouterJobStatus,
    RouterWorker,
    RouterJobOffer,
    AcceptJobOfferResult,
    UnassignJobResult,
    RouterJob,
    RouterJobAssignment,
    RouterWorkerState,
    DistributionPolicy,
    RouterQueue,
    DeclineJobOfferOptions,
    CompleteJobOptions,
    CloseJobOptions,
)

channel_id = "fakeChannel2"


# The test class name needs to start with "Test" to get collected by pytest
class TestLimitConcurrentOffersToWorkerScenario(RouterRecordedTestCase):
    def clean_up(self, **kwargs):
        # delete in live mode
        if not self.is_playback():
            router_admin_client: JobRouterAdministrationClient = self.create_admin_client()
            router_client: JobRouterClient = self.create_client()

            if self._testMethodName in self.job_ids and any(self.job_ids[self._testMethodName]):
                for _id in set(self.job_ids[self._testMethodName]):
                    self.clean_up_job(job_id=_id, suppress_errors=True)

            if self._testMethodName in self.worker_ids and any(self.worker_ids[self._testMethodName]):
                for _id in set(self.worker_ids[self._testMethodName]):
                    # delete worker
                    router_client.delete_worker(_id)

            if self._testMethodName in self.classification_policy_ids and any(
                self.classification_policy_ids[self._testMethodName]
            ):
                for policy_id in set(self.classification_policy_ids[self._testMethodName]):
                    router_admin_client.delete_classification_policy(policy_id)

            if self._testMethodName in self.queue_ids and any(self.queue_ids[self._testMethodName]):
                for policy_id in set(self.queue_ids[self._testMethodName]):
                    router_admin_client.delete_queue(policy_id)

            if self._testMethodName in self.distribution_policy_ids and any(
                self.distribution_policy_ids[self._testMethodName]
            ):
                for policy_id in set(self.distribution_policy_ids[self._testMethodName]):
                    router_admin_client.delete_distribution_policy(policy_id)

    def get_distribution_policy_id(self, **kwargs):
        return "_" + self._testMethodName + "_tst_dp"

    def setup_distribution_policy(self, **kwargs):
        client: JobRouterAdministrationClient = self.create_admin_client()
        distribution_policy_id = self.get_distribution_policy_id()

        policy: DistributionPolicy = DistributionPolicy(
            offer_expires_after_seconds=10.0 * 60,
            mode=LongestIdleMode(min_concurrent_offers=1, max_concurrent_offers=1),
            name="test",
        )

        distribution_policy = client.upsert_distribution_policy(distribution_policy_id, policy)

        # add for cleanup later
        if self._testMethodName in self.distribution_policy_ids:
            self.distribution_policy_ids[self._testMethodName] = self.distribution_policy_ids[
                self._testMethodName
            ].append(distribution_policy_id)
        else:
            self.distribution_policy_ids[self._testMethodName] = [distribution_policy_id]

    def get_job_queue_id(self, **kwargs):
        return "_" + self._testMethodName + "_tst_q"

    def setup_job_queue(self, **kwargs):
        client: JobRouterAdministrationClient = self.create_admin_client()
        job_queue_id = self.get_job_queue_id()

        job_queue: RouterQueue = RouterQueue(distribution_policy_id=self.get_distribution_policy_id(), name="test")

        job_queue = client.upsert_queue(job_queue_id, job_queue)

        # add for cleanup later
        if self._testMethodName in self.queue_ids:
            self.queue_ids[self._testMethodName] = self.queue_ids[self._testMethodName].append(job_queue_id)
        else:
            self.queue_ids[self._testMethodName] = [job_queue_id]

    def get_router_worker_id(self, **kwargs):
        return "_" + self._testMethodName + "_tst_w"

    def setup_router_worker(self, **kwargs):
        w_identifier = self.get_router_worker_id()
        router_client: JobRouterClient = self.create_client()
        worker_queues = [self.get_job_queue_id()]
        worker_channels = [RouterChannel(channel_id=channel_id, capacity_cost_per_job=1)]

        worker: RouterWorker = RouterWorker(
            capacity=10,
            queues=worker_queues,
            channels=worker_channels,
            available_for_offers=True,
            max_concurrent_offers=1,
        )

        router_worker = router_client.upsert_worker(w_identifier, worker)

        # add for cleanup
        self.worker_ids[self._testMethodName] = [w_identifier]

    def validate_job_is_queued(self, identifier, **kwargs):
        router_client: JobRouterClient = self.create_client()
        router_job = router_client.get_job(identifier)
        assert router_job.status == RouterJobStatus.QUEUED

    def validate_worker_has_offer(
        self,
        worker_id,  # type: str
        job_ids,  # type: List[str]
        **kwargs,  # type: Any
    ):
        router_client: JobRouterClient = self.create_client()
        router_worker: RouterWorker = router_client.get_worker(worker_id)
        offer_for_jobs = [job_offer for job_offer in router_worker.offers if job_offer.job_id in job_ids]
        assert any(offer_for_jobs)

    def validate_worker_deregistration(  # cSpell:disable-line
        self,
        worker_id,  # type: str
        **kwargs,  # type: Any
    ):
        router_client: JobRouterClient = self.create_client()
        router_worker: RouterWorker = router_client.get_worker(worker_id)
        assert router_worker.state == RouterWorkerState.INACTIVE

    @RouterPreparers.router_test_decorator
    @recorded_by_proxy
    @RouterPreparers.before_test_execute("setup_distribution_policy")
    @RouterPreparers.before_test_execute("setup_job_queue")
    @RouterPreparers.before_test_execute("setup_router_worker")
    @RouterPreparers.after_test_execute("clean_up")
    def test_limit_concurrent_offer_to_worker_scenario(self, **kwargs):
        router_client: JobRouterClient = self.create_client()

        # create job
        job_identifier_1 = f"concurrent_offer_worker_scenario_job_id_1"
        job_identifier_2 = f"concurrent_offer_worker_scenario_job_id_2"
        job_ids = [job_identifier_1, job_identifier_2]

        router_job_1: RouterJob = RouterJob(
            channel_id=channel_id,
            queue_id=self.get_job_queue_id(),
            priority=1,
        )
        router_job_1: RouterJob = router_client.upsert_job(job_identifier_1, router_job_1)

        # add for cleanup
        self.job_ids[self._testMethodName] = [job_identifier_1]

        assert router_job_1 is not None

        # validate there are offers made
        self._poll_until_no_exception(
            self.validate_worker_has_offer, AssertionError, self.get_router_worker_id(), job_ids
        )

        router_job_2: RouterJob = RouterJob(
            channel_id=channel_id,
            queue_id=self.get_job_queue_id(),
            priority=1,
        )
        router_job_2: RouterJob = router_client.upsert_job(job_identifier_2, router_job_2)

        # add for cleanup
        self.job_ids[self._testMethodName].append(job_identifier_2)

        assert router_job_2 is not None

        # validate there are still offers
        self._poll_until_no_exception(
            self.validate_worker_has_offer, AssertionError, self.get_router_worker_id(), job_ids
        )

        router_worker = router_client.get_worker(self.get_router_worker_id())
        job_offers = [job_offer for job_offer in router_worker.offers if job_offer.job_id in job_ids]

        assert len(job_offers) == 1
        job_offer_1: RouterJobOffer = job_offers[0]
        assert job_offer_1.capacity_cost == 1
        assert job_offer_1.offered_at is not None
        assert job_offer_1.expires_at is not None
        job_without_offer: str = job_identifier_2 if job_offer_1.job_id == job_identifier_1 else job_identifier_1

        # accept job offer
        offer_id_1: str = job_offer_1.offer_id
        accept_job_offer_result_1: AcceptJobOfferResult = router_client.accept_job_offer(
            worker_id=self.get_router_worker_id(), offer_id=offer_id_1
        )

        assert accept_job_offer_result_1.job_id == job_offer_1.job_id
        assert accept_job_offer_result_1.worker_id == self.get_router_worker_id()

        # check for new offers sent to worker now that previous offer has been accepted
        self._poll_until_no_exception(
            self.validate_worker_has_offer, AssertionError, self.get_router_worker_id(), [job_without_offer]
        )

        router_worker = router_client.get_worker(self.get_router_worker_id())
        job_offers = [job_offer for job_offer in router_worker.offers if job_offer.job_id == job_without_offer]

        assert len(job_offers) == 1
        job_offer_2: RouterJobOffer = job_offers[0]
        assert job_offer_2.capacity_cost == 1
        assert job_offer_2.offered_at is not None
        assert job_offer_2.expires_at is not None

        # accept job offer
        offer_id_2 = job_offer_2.offer_id
        accept_job_offer_result_2: AcceptJobOfferResult = router_client.accept_job_offer(
            worker_id=self.get_router_worker_id(), offer_id=offer_id_2
        )

        assert accept_job_offer_result_2.job_id == job_without_offer
        assert accept_job_offer_result_2.worker_id == self.get_router_worker_id()

        # complete both jobs
        router_client.complete_job(accept_job_offer_result_1.job_id, accept_job_offer_result_1.assignment_id)
        router_client.complete_job(accept_job_offer_result_2.job_id, accept_job_offer_result_2.assignment_id)

        # close both jobs
        router_client.close_job(accept_job_offer_result_1.job_id, accept_job_offer_result_1.assignment_id)
        router_client.close_job(accept_job_offer_result_2.job_id, accept_job_offer_result_2.assignment_id)

        # validate post closure job details
        queried_job_1: RouterJob = router_client.get_job(job_identifier_1)
        queried_job_2: RouterJob = router_client.get_job(job_identifier_2)

        assignment_id_for_job_1 = (
            accept_job_offer_result_1.assignment_id
            if accept_job_offer_result_1.job_id == job_identifier_1
            else accept_job_offer_result_2.assignment_id
        )
        assignment_id_for_job_2 = (
            accept_job_offer_result_1.assignment_id
            if accept_job_offer_result_1.job_id == job_identifier_2
            else accept_job_offer_result_2.assignment_id
        )

        job_assignment_1: RouterJobAssignment = queried_job_1.assignments[assignment_id_for_job_1]
        assert job_assignment_1.assigned_at is not None
        assert job_assignment_1.worker_id == self.get_router_worker_id()
        assert job_assignment_1.completed_at is not None
        assert job_assignment_1.closed_at is not None

        job_assignment_2: RouterJobAssignment = queried_job_2.assignments[assignment_id_for_job_2]
        assert job_assignment_2.assigned_at is not None
        assert job_assignment_2.worker_id == self.get_router_worker_id()
        assert job_assignment_2.completed_at is not None
        assert job_assignment_2.closed_at is not None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
FILE: router_worker_crud_ops.py
DESCRIPTION:
    These samples demonstrates how to create Workers used in ACS JobRouter.
    You need a valid connection string to an Azure Communication Service to execute the sample

USAGE:
    python router_worker_crud_ops.py
    Set the environment variables with your own values before running the sample:
    1) AZURE_COMMUNICATION_SERVICE_CONNECTION_STRING - Communication Service connection string
"""

import os


class RouterWorkerSamples(object):
    connection_string = os.environ["AZURE_COMMUNICATION_SERVICE_CONNECTION_STRING"]

    _worker_id = "sample_worker"
    _distribution_policy_id = "sample_dp_policy"

    def setup_distribution_policy(self):
        connection_string = self.connection_string
        distribution_policy_id = self._distribution_policy_id

        from azure.communication.jobrouter import JobRouterAdministrationClient
        from azure.communication.jobrouter.models import LongestIdleMode, DistributionPolicy

        router_admin_client = JobRouterAdministrationClient.from_connection_string(conn_str=connection_string)
        distribution_policy = router_admin_client.upsert_distribution_policy(
            distribution_policy_id,
            DistributionPolicy(
                offer_expires_after_seconds=10 * 60,
                mode=LongestIdleMode(min_concurrent_offers=1, max_concurrent_offers=1),
            ),
        )
        print(f"Sample setup completed: Created distribution policy")

    def setup_queues(self):
        connection_string = self.connection_string
        distribution_policy_id = self._distribution_policy_id

        from azure.communication.jobrouter import JobRouterAdministrationClient
        from azure.communication.jobrouter.models import RouterQueue

        router_admin_client = JobRouterAdministrationClient.from_connection_string(conn_str=connection_string)
        job_queue1: RouterQueue = router_admin_client.upsert_queue(
            "worker-q-1",
            RouterQueue(
                distribution_policy_id=distribution_policy_id,
            ),
        )

        job_queue2: RouterQueue = router_admin_client.upsert_queue(
            "worker-q-2",
            RouterQueue(
                distribution_policy_id=distribution_policy_id,
            ),
        )

        job_queue3: RouterQueue = router_admin_client.upsert_queue(
            "worker-q-3",
            RouterQueue(
                distribution_policy_id=distribution_policy_id,
            ),
        )

        print(f"Sample setup completed: Created queues")

    def create_worker(self):
        connection_string = self.connection_string
        worker_id = self._worker_id
        # [START create_worker]
        from azure.communication.jobrouter import (
            JobRouterClient,
        )
        from azure.communication.jobrouter.models import (
            RouterWorker,
            RouterChannel,
        )

        # set `connection_string` to an existing ACS endpoint
        router_client = JobRouterClient.from_connection_string(conn_str=connection_string)
        print("JobRouterClient created successfully!")

        router_worker: RouterWorker = router_client.upsert_worker(
            worker_id,
            RouterWorker(
                capacity=100,
                queues=["worker-q-1", "worker-q-2"],
                channels=[
                    RouterChannel(channel_id="WebChat", capacity_cost_per_job=1),
                    RouterChannel(channel_id="WebChatEscalated", capacity_cost_per_job=20),
                    RouterChannel(channel_id="Voip", capacity_cost_per_job=100),
                ],
                labels={"Location": "NA", "English": 7, "O365": True, "Xbox_Support": False},
                tags={"Name": "John Doe", "Department": "IT_HelpDesk"},
            ),
        )

        print(f"Router worker successfully created with id: {router_worker.id}")

        # [END create_worker]

    def create_worker_w_limit_concurrent_offers(self):
        connection_string = self.endpoint
        worker_id = self._worker_id
        # [START create_worker_w_limit_concurrent_offers]
        from azure.communication.jobrouter import (
            JobRouterClient,
        )
        from azure.communication.jobrouter.models import (
            RouterWorker,
            RouterChannel,
        )

        # set `connection_string` to an existing ACS endpoint
        router_client = JobRouterClient.from_connection_string(conn_str=connection_string)
        print("JobRouterClient created successfully!")

        router_worker: RouterWorker = router_client.upsert_worker(
            worker_id,
            RouterWorker(
                capacity=100,
                queues=["worker-q-1", "worker-q-2"],
                channels=[
                    RouterChannel(channel_id="WebChat", capacity_cost_per_job=1),
                    RouterChannel(channel_id="WebChatEscalated", capacity_cost_per_job=20),
                    RouterChannel(channel_id="Voip", capacity_cost_per_job=100),
                ],
                labels={"Location": "NA", "English": 7, "O365": True, "Xbox_Support": False},
                tags={"Name": "John Doe", "Department": "IT_HelpDesk"},
                max_concurrent_offers=1,
            ),
        )

        print(f"Router worker successfully created with id: {router_worker.id}")

        # [END create_worker_w_limit_concurrent_offers]

    def update_worker(self):
        connection_string = self.connection_string
        worker_id = self._worker_id
        # [START update_worker]
        from azure.communication.jobrouter import (
            JobRouterClient,
        )
        from azure.communication.jobrouter.models import (
            RouterWorker,
            RouterChannel,
        )

        # set `connection_string` to an existing ACS endpoint
        router_client: JobRouterClient = JobRouterClient.from_connection_string(conn_str=connection_string)
        print("JobRouterClient created successfully!")

        # we are going to
        # 1. Assign the worker to another queue
        # 2. Modify an value of label: `O365`
        # 3. Delete label: `Xbox_Support`
        # 4. Add a new label: `Xbox_Support_EN` and set value true
        # 5. Increase capacityCostPerJob for channel `WebChatEscalated` to 50

        updated_router_worker: RouterWorker = router_client.upsert_worker(
            worker_id,
            queues=["worker-q-3"],
            channels=[RouterChannel(channel_id="WebChatEscalated", capacity_cost_per_job=50)],
            labels={"O365": "Supported", "Xbox_Support": None, "Xbox_Support_EN": True},
        )

        print(f"Router worker successfully update with labels {updated_router_worker.labels}")
        # [END update_worker]

    def get_worker(self):
        connection_string = self.connection_string
        worker_id = self._worker_id
        # [START get_worker]
        from azure.communication.jobrouter import JobRouterClient

        router_client = JobRouterClient.from_connection_string(conn_str=connection_string)

        router_worker = router_client.get_worker(worker_id=worker_id)

        print(f"Successfully fetched router worker with id: {router_worker.id}")
        # [END get_worker]

    def register_worker(self):
        connection_string = self.connection_string
        worker_id = self._worker_id
        # [START register_worker]
        from azure.communication.jobrouter import JobRouterClient

        router_client = JobRouterClient.from_connection_string(conn_str=connection_string)

        router_worker = router_client.upsert_worker(worker_id, available_for_offers=True)

        print(f"Successfully registered router worker with id: {router_worker.id} with status: {router_worker.state}")
        # [END register_worker]

    def deregister_worker(self):
        connection_string = self.connection_string
        worker_id = self._worker_id
        # [START deregister_worker]
        from azure.communication.jobrouter import JobRouterClient

        router_client = JobRouterClient.from_connection_string(conn_str=connection_string)

        router_worker = router_client.upsert_worker(worker_id, available_for_offers=False)

        print(
            f"Successfully de-registered router worker with id: {router_worker.id} with status: {router_worker.state}"
        )
        # [END deregister_worker]

    def list_workers(self):
        connection_string = self.connection_string
        # [START list_workers]
        from azure.communication.jobrouter import JobRouterClient

        router_client = JobRouterClient.from_connection_string(conn_str=connection_string)

        router_worker_iterator = router_client.list_workers()

        for w in router_worker_iterator:
            print(f"Retrieved worker with id: {w.id}")

        print(f"Successfully completed fetching workers")
        # [END list_workers]

    def list_workers_batched(self):
        connection_string = self.connection_string
        # [START list_workers_batched]
        from azure.communication.jobrouter import JobRouterClient

        router_client = JobRouterClient.from_connection_string(conn_str=connection_string)

        router_worker_iterator = router_client.list_workers(results_per_page=10)

        for worker_page in router_worker_iterator.by_page():
            workers_in_page = list(worker_page)
            print(f"Retrieved {len(workers_in_page)} workers in current page")

            for w in workers_in_page:
                print(f"Retrieved worker with id: {w.id}")

        print(f"Successfully completed fetching workers")
        # [END list_workers_batched]

    def clean_up(self):
        connection_string = self.connection_string
        worker_id = self._worker_id

        # [START delete_worker]
        from azure.communication.jobrouter import JobRouterClient

        router_client = JobRouterClient.from_connection_string(conn_str=connection_string)

        router_client.delete_worker(worker_id=worker_id)

        # [END delete_worker]


if __name__ == "__main__":
    sample = RouterWorkerSamples()
    sample.setup_distribution_policy()
    sample.setup_queues()
    sample.create_worker()
    sample.create_worker_w_limit_concurrent_offers()
    sample.update_worker()
    sample.get_worker()
    sample.register_worker()
    sample.deregister_worker()
    sample.list_workers()
    sample.list_workers_batched()
    sample.clean_up()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
"""Schema walking helpers for validator generation."""

from __future__ import annotations

from typing import Any


def resolve_ref(ref: str) -> str:
    """Extract schema name from OpenAPI $ref values."""
    return ref.rsplit("/", 1)[-1]


def _iter_subschemas(schema: dict[str, Any]) -> list[dict[str, Any]]:
    """Yield nested schema objects that may contain references."""
    nested: list[dict[str, Any]] = []

    for key in ("oneOf", "anyOf", "allOf"):
        branches = schema.get(key, [])
        if isinstance(branches, list):
            nested.extend([branch for branch in branches if isinstance(branch, dict)])

    properties = schema.get("properties", {})
    if isinstance(properties, dict):
        nested.extend([value for value in properties.values() if isinstance(value, dict)])

    items = schema.get("items")
    if isinstance(items, dict):
        nested.append(items)

    additional = schema.get("additionalProperties")
    if isinstance(additional, dict):
        nested.append(additional)

    # Walk discriminator mapping refs so schemas like ItemMessage / FunctionTool
    # that are only reachable via a discriminator are included in the reachable set
    # (and therefore have the overlay applied before the emitter sees them).
    disc = schema.get("discriminator")
    if isinstance(disc, dict):
        mapping = disc.get("mapping", {})
        if isinstance(mapping, dict):
            for ref_str in mapping.values():
                if isinstance(ref_str, str):
                    nested.append({"$ref": ref_str})

    return nested


class SchemaWalker:
    """Collect schemas reachable from one or more roots."""

    def __init__(
        self,
        schemas: dict[str, dict[str, Any]],
        overlay: dict[str, Any] | None = None,
    ) -> None:
        self.schemas = schemas
        self.overlay: dict[str, Any] = overlay or {}
        self.reachable: dict[str, dict[str, Any]] = {}
        self._visited: set[str] = set()

    def walk(self, name: str) -> None:
        """Walk a schema by name and recursively collect reachable references."""
        if name in self._visited:
            return
        self._visited.add(name)

        schema = self.schemas.get(name)
        if schema is None:
            return

        schema = self._apply_overlay(name, dict(schema))
        self.reachable[name] = schema
        self._walk_schema(schema)

    def _apply_overlay(self, name: str, schema: dict[str, Any]) -> dict[str, Any]:
        """Apply overlay fixes to a schema.

        Supports four overlay keys per schema entry:
        - ``required``: replace the required list entirely.
        - ``not_required``: remove individual fields from required and mark their
          property schemas as ``nullable`` so ``None`` is accepted.
        - ``properties``: merge per-property constraint overrides (e.g. minimum/maximum).
        - ``default_discriminator``: set a default discriminator value for the schema's
          discriminator dispatch. When the discriminator property is absent from the
          payload, this value is used instead of rejecting the input.
        """
        overlay_schemas = self.overlay.get("schemas", {})
        # Try exact name first, then fall back to the name with any "Vendor." prefix stripped
        # (e.g. "OpenAI.ItemMessage" -> "ItemMessage") to stay compatible with the overlay
        # format shared with the TypeSpec code generator, where TypeSpec uses bare names.
        overlay_entry = overlay_schemas.get(name)
        if not overlay_entry:
            bare = name.rsplit(".", 1)[-1] if "." in name else None
            if bare:
                overlay_entry = overlay_schemas.get(bare)
        if not overlay_entry:
            return schema

        # Replace required list entirely
        if "required" in overlay_entry:
            schema["required"] = list(overlay_entry["required"])

        # Remove individual fields from required; mark those properties nullable
        if "not_required" in overlay_entry:
            current_required = list(schema.get("required", []))
            for field in overlay_entry["not_required"]:
                if field in current_required:
                    current_required.remove(field)
                # Mark property nullable so the emitter accepts None/absent values
                props = schema.get("properties")
                if isinstance(props, dict) and field in props:
                    props[field] = dict(props[field])
                    props[field]["nullable"] = True
            schema["required"] = current_required

        # Merge property-level constraint overrides
        if "properties" in overlay_entry:
            if "properties" not in schema:
                schema["properties"] = {}
            for prop_name, constraints in overlay_entry["properties"].items():
                if prop_name not in schema["properties"]:
                    schema["properties"][prop_name] = {}
                else:
                    schema["properties"][prop_name] = dict(schema["properties"][prop_name])
                schema["properties"][prop_name].update(constraints)

        # Inject default_discriminator into the schema's discriminator dict
        if "default_discriminator" in overlay_entry:
            disc = schema.get("discriminator")
            if isinstance(disc, dict):
                schema["discriminator"] = dict(disc)
                schema["discriminator"]["defaultValue"] = overlay_entry["default_discriminator"]

        return schema

    def _walk_schema(self, schema: dict[str, Any]) -> None:
        """Walk nested schema branches."""
        ref = schema.get("$ref")
        if isinstance(ref, str):
            self.walk(resolve_ref(ref))
            return

        for nested in _iter_subschemas(schema):
            self._walk_schema(nested)


def discover_post_request_roots(spec: dict[str, Any]) -> list[str]:
    """Discover root schema names referenced by POST request bodies."""
    roots: list[str] = []
    paths = spec.get("paths", {})

    for _path, methods in sorted(paths.items()):
        if not isinstance(methods, dict):
            continue
        post = methods.get("post")
        if not isinstance(post, dict):
            continue
        request_body = post.get("requestBody", {})
        content = request_body.get("content", {}).get("application/json", {})
        schema = content.get("schema", {})

        if isinstance(schema, dict) and isinstance(schema.get("$ref"), str):
            roots.append(resolve_ref(schema["$ref"]))
            continue

        if isinstance(schema, dict):
            for key in ("oneOf", "anyOf"):
                branches = schema.get(key, [])
                if not isinstance(branches, list):
                    continue
                for branch in branches:
                    if isinstance(branch, dict) and isinstance(branch.get("$ref"), str):
                        roots.append(resolve_ref(branch["$ref"]))

    deduped: list[str] = []
    seen: set[str] = set()
    for root in roots:
        if root not in seen:
            seen.add(root)
            deduped.append(root)
    return deduped



# pylint: disable=line-too-long,useless-suppression
import functools
from devtools_testutils import AzureRecordedTestCase, PowerShellPreparer, recorded_by_proxy
from azure.core.credentials import AzureKeyCredential
from azure.ai.textanalytics.authoring import TextAuthoringClient
from azure.ai.textanalytics.authoring.models import (
    CreateProjectOptions,
    ProjectDetails,
    ProjectKind,
)

ConversationsPreparer = functools.partial(
    PowerShellPreparer,
    "authoring",
    authoring_endpoint="https://Sanitized.cognitiveservices.azure.com/",
    authoring_key="fake_key",
)


class TestConversations(AzureRecordedTestCase):
    def create_client(self, endpoint: str, key: str) -> TextAuthoringClient:
        return TextAuthoringClient(endpoint, AzureKeyCredential(key))


class TestConversationsCase(TestConversations):
    @ConversationsPreparer()
    @recorded_by_proxy
    def test_create_project(self, authoring_endpoint, authoring_key):
        authoring_client = self.create_client(authoring_endpoint, authoring_key)
        project_name = "MyTextProject0902"

        # Arrange
        body = CreateProjectOptions(
            project_kind=ProjectKind.CUSTOM_MULTI_LABEL_CLASSIFICATION,
            storage_input_container_name="multi-class-example",
            project_name=project_name,
            language="en",
            description="Project description for a Custom Entity Recognition project",
            multilingual=True,
        )

        # Act
        result = authoring_client.create_project(project_name=project_name, body=body)

        # Assert
        assert isinstance(result, ProjectDetails)
        assert result.project_name == project_name
        print(f"Created project: {result.project_name}, language: {result.language}, kind: {result.project_kind}")



from contextlib import suppress
import logging
import os
from pathlib import Path
from typing import Dict, Any, Optional, List

from jinja2 import Template, PackageLoader, Environment
from .conf import read_conf, build_default_conf, CONF_NAME

_LOGGER = logging.getLogger(__name__)

_CWD = Path(__file__).resolve().parent
_TEMPLATE_PATH = _CWD / "template"


def build_config(config: Dict[str, Any]) -> Dict[str, str]:
    """Will build the actual config for Jinja2, based on SDK config."""
    result = config.copy()
    # Manage the classifier stable/beta
    is_stable = result.pop("is_stable", False)
    if is_stable:
        result["classifier"] = "Development Status :: 5 - Production/Stable"
    else:
        result["classifier"] = "Development Status :: 4 - Beta"
    # Manage the nspkg
    package_name = result["package_name"]
    result["package_nspkg"] = result.pop("package_nspkg", package_name[: package_name.rindex("-")] + "-nspkg")
    # ARM?
    result["is_arm"] = result.pop("is_arm", True)

    # Do I need msrestazure for this package?
    result["need_msrestazure"] = result.pop("need_msrestazure", False)

    # Do I need azure-mgmt-core for this package?
    result["need_azuremgmtcore"] = result.pop("need_azuremgmtcore", True)

    # Pre-compute some Jinja variable that are complicated to do inside the templates
    package_parts = result["package_nspkg"][: -len("-nspkg")].split("-")
    result["nspkg_names"] = [".".join(package_parts[: i + 1]) for i in range(len(package_parts))]
    result["init_names"] = ["/".join(package_parts[: i + 1]) + "/__init__.py" for i in range(len(package_parts))]
    exclude_folders = result.pop("exclude_folders", "")
    result["exclude_folders"] = [item.strip() for item in exclude_folders.split(",") if item.strip()]

    # Return result
    return result


def build_packaging(
    output_folder: str,
    gh_token: Optional[str] = None,
    jenkins: bool = False,
    packages: List[str] = None,
    build_conf: bool = False,
    *,
    template_names: Optional[List[str]] = None,
) -> None:
    package_names = set(packages) or set()
    if jenkins:
        sdk_id = os.environ["ghprbGhRepository"]
        pr_number = int(os.environ["ghprbPullId"])

        from github import Github

        con = Github(gh_token)
        repo = con.get_repo(sdk_id)
        sdk_pr = repo.get_pull(pr_number)
        # "get_files" of Github only download the first 300 files. Might not be enough.
        package_names |= {f.filename.split("/")[0] for f in sdk_pr.get_files() if f.filename.startswith("azure")}

    if not package_names:
        raise ValueError("Was unable to find out the package names.")

    for package_name in package_names:
        build_packaging_by_package_name(package_name, output_folder, build_conf, template_names)


def build_packaging_by_package_name(
    package_name: str, output_folder: str, build_conf: bool = False, template_names: Optional[List[str]] = None
) -> None:
    _LOGGER.info("Building template %s", package_name)
    package_folder = Path(output_folder) / Path(package_name)

    if build_conf:
        build_default_conf(package_folder, package_name)

    conf = read_conf(package_folder)
    if not conf:
        raise ValueError("Create a {} file before calling this script".format(package_folder / CONF_NAME))

    if not conf.get("auto_update", True):
        _LOGGER.info(f"Package {package_name} has no auto-packaging update enabled")
        return

    env = Environment(loader=PackageLoader("packaging_tools", "templates/packaging_files"), keep_trailing_newline=True)
    conf = build_config(conf)

    template_names = template_names or env.list_templates()
    for template_name in env.list_templates():
        if template_name not in template_names:
            _LOGGER.info("Skipping template %s", template_name)
            continue
        future_filepath = Path(output_folder) / package_name / template_name

        # Might decide to make it more generic one day
        if template_name == "CHANGELOG.md" and future_filepath.exists():
            _LOGGER.info("Skipping CHANGELOG.md template, since a previous one was found")
            # Never overwirte the ChangeLog
            continue

        template = env.get_template(template_name)
        result = template.render(**conf)

        # __init__.py is a weird one
        if template_name == "__init__.py":
            split_package_name = package_name.split("-")[:-1]
            for i in range(len(split_package_name)):
                init_path = Path(output_folder).joinpath(package_name, *split_package_name[: i + 1], template_name)
                with open(init_path, "w") as fd:
                    fd.write(result)

            continue

        with open(future_filepath, "w") as fd:
            fd.write(result)
    # azure_bdist_wheel had been removed, but need to delete it manually
    with suppress(FileNotFoundError):
        (Path(output_folder) / package_name / "azure_bdist_wheel.py").unlink()

    _LOGGER.info("Template done %s", package_name)



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
import pytest
from azure.mgmt.imagebuilder.aio import ImageBuilderClient

from devtools_testutils import AzureMgmtRecordedTestCase, RandomNameResourceGroupPreparer
from devtools_testutils.aio import recorded_by_proxy_async

AZURE_LOCATION = "eastus"


@pytest.mark.skip("you may need to update the auto-generated test case before run it")
class TestImageBuilderVirtualMachineImageTemplatesOperationsAsync(AzureMgmtRecordedTestCase):
    def setup_method(self, method):
        self.client = self.create_mgmt_client(ImageBuilderClient, is_async=True)

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_virtual_machine_image_templates_get(self, resource_group):
        response = await self.client.virtual_machine_image_templates.get(
            resource_group_name=resource_group.name,
            image_template_name="str",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_virtual_machine_image_templates_begin_create_or_update(self, resource_group):
        response = await (
            await self.client.virtual_machine_image_templates.begin_create_or_update(
                resource_group_name=resource_group.name,
                image_template_name="str",
                parameters={
                    "identity": {
                        "type": "str",
                        "userAssignedIdentities": {"str": {"clientId": "str", "principalId": "str"}},
                    },
                    "location": "str",
                    "id": "str",
                    "name": "str",
                    "properties": {
                        "distribute": ["image_template_distributor"],
                        "source": "image_template_source",
                        "additionalDataDisks": [{"sizeGB": 0}],
                        "autoRun": {"state": "str"},
                        "buildTimeoutInMinutes": 0,
                        "customize": ["image_template_customizer"],
                        "errorHandling": {"onCustomizerError": "str", "onValidationError": "str"},
                        "exactStagingResourceGroup": "str",
                        "lastRunStatus": {
                            "endTime": "2020-02-20 00:00:00",
                            "message": "str",
                            "runState": "str",
                            "runSubState": "str",
                            "startTime": "2020-02-20 00:00:00",
                        },
                        "managedResourceTags": {"str": "str"},
                        "optimize": {
                            "vmBoot": {"state": "str"},
                            "workload": {"scriptUri": "str", "sha256Checksum": "str", "state": "str"},
                        },
                        "provisioningError": {"message": "str", "provisioningErrorCode": "str"},
                        "provisioningState": "str",
                        "stagingResourceGroup": "str",
                        "validate": {
                            "continueDistributeOnFailure": bool,
                            "inVMValidations": ["image_template_in_vm_validator"],
                            "sourceValidationOnly": bool,
                        },
                        "vmProfile": {
                            "osDiskSizeGB": 0,
                            "userAssignedIdentities": ["str"],
                            "vmSize": "str",
                            "vnetConfig": {"containerInstanceSubnetId": "str", "proxyVmSize": "str", "subnetId": "str"},
                        },
                    },
                    "systemData": {
                        "createdAt": "2020-02-20 00:00:00",
                        "createdBy": "str",
                        "createdByType": "str",
                        "lastModifiedAt": "2020-02-20 00:00:00",
                        "lastModifiedBy": "str",
                        "lastModifiedByType": "str",
                    },
                    "tags": {"str": "str"},
                    "type": "str",
                },
            )
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_virtual_machine_image_templates_begin_update(self, resource_group):
        response = await (
            await self.client.virtual_machine_image_templates.begin_update(
                resource_group_name=resource_group.name,
                image_template_name="str",
                parameters={
                    "identity": {
                        "type": "str",
                        "userAssignedIdentities": {"str": {"clientId": "str", "principalId": "str"}},
                    },
                    "properties": {
                        "distribute": ["image_template_distributor"],
                        "vmProfile": {
                            "osDiskSizeGB": 0,
                            "userAssignedIdentities": ["str"],
                            "vmSize": "str",
                            "vnetConfig": {"containerInstanceSubnetId": "str", "proxyVmSize": "str", "subnetId": "str"},
                        },
                    },
                    "tags": {"str": "str"},
                },
            )
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_virtual_machine_image_templates_begin_delete(self, resource_group):
        response = await (
            await self.client.virtual_machine_image_templates.begin_delete(
                resource_group_name=resource_group.name,
                image_template_name="str",
            )
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_virtual_machine_image_templates_list_by_resource_group(self, resource_group):
        response = self.client.virtual_machine_image_templates.list_by_resource_group(
            resource_group_name=resource_group.name,
        )
        result = [r async for r in response]
        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_virtual_machine_image_templates_list(self, resource_group):
        response = self.client.virtual_machine_image_templates.list()
        result = [r async for r in response]
        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_virtual_machine_image_templates_begin_run(self, resource_group):
        response = await (
            await self.client.virtual_machine_image_templates.begin_run(
                resource_group_name=resource_group.name,
                image_template_name="str",
            )
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_virtual_machine_image_templates_begin_cancel(self, resource_group):
        response = await (
            await self.client.virtual_machine_image_templates.begin_cancel(
                resource_group_name=resource_group.name,
                image_template_name="str",
            )
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_virtual_machine_image_templates_get_run_output(self, resource_group):
        response = await self.client.virtual_machine_image_templates.get_run_output(
            resource_group_name=resource_group.name,
            image_template_name="str",
            run_output_name="str",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_virtual_machine_image_templates_list_run_outputs(self, resource_group):
        response = self.client.virtual_machine_image_templates.list_run_outputs(
            resource_group_name=resource_group.name,
            image_template_name="str",
        )
        result = [r async for r in response]
        # please add some check logic here by yourself
        # ...
